#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from ..helpers import keysToCamelNested, getPropertyPath, isPropertyPathCorrect, isMethodPathCorrect, getMethodPath, \
    getEmitterPath, isEmitterPathCorrect, ensureDottedAccess, EXECUTOR

from ..logger import getNopeLogger


class BaseModule(object):
    """A Base Class Describing a Module.
    """

    decoratedItems = list()

    @property
    def type(self):
        """Type Idefintifier of the Element

        Returns:
            str: Type Idefintifier of the Element
        """
        return self._type

    def setAuthor(self, forename: str, surename: str, mail: str):
        """Custom Helper Function to generate an author

        Args:
            forename (str): Forename of the Author
            surename (str): Surename of the Author
            mail (str): Mail of the Author
        """
        self._author = {
            'forename': forename,
            'surename': surename,
            'mail': mail
        }

    @property
    def author(self):
        """Helper to return the Author

        Returns:
            dict: A Dict, describing the Author
        """
        return self._author

    def setVersion(self, version: str, date):
        """Helper Function to set a Version of a Module.

        Args:
            version (str): The Version as String. E.G. 0.9.0
            date (date): The Date of the Version
        """
        self._version = {
            'version': version,
            'date': date
        }

    @property
    def version(self):
        """Getter for the Version

        Returns:
            dict: A Dict describing the Version
        """
        return self._version

    identifier: str
    description: str

    @property
    def methods(self) -> dict:
        """Helper to list the Functions

        Returns:
            dict: A List with all Functions available.
        """
        ret = {}

        for name, funcs in self._registeredMethods.items():
            ret[name] = funcs["options"]

        return ret

    @property
    def properties(self) -> dict:
        """Helper to list all Properties.

        Returns:
            dict: [description]
        """

        ret = {}

        for name, prop in self._registeredProperties.items():
            ret[name] = prop["options"]
        return ret

    @property
    def events(self) -> dict:
        """Helper to list all Properties.

        Returns:
            dict: [description]
        """

        ret = {}

        for name, prop in self._registeredEvents.items():
            ret[name] = prop["options"]
        return ret

    def __init__(self, core):
        super().__init__()

        if not hasattr(self, "decoratedItems"):
            setattr(self, "decoratedItems", [])

        self._logger = getNopeLogger(self.__class__.__name__)

        self._core = core

        self.description = None
        self._author = None
        self._version = None
        self.identifier = None

        self._registeredMethods = dict()
        self._registeredProperties = dict()
        self._registeredEvents = dict()
        self._type = str(self.__class__.__name__)
        self.uiLinks = []

        # Helper for the Decorator
        self._markedElements = dict()

    async def registerProperty(self, name: str, observable, options=None):
        """ Function to register a property. This will provide the property in the NoPE-Environment.

        Args:
            name (str): Name of the property, which should be used to forward the data
            observable (nope-observable):  the property to use
            options (dict): options, to assign a different topic and options to select, whether the content of the property should be "published" and/or "subscribed"

        Returns:
            [type]: [description]
        """

        options = ensureDottedAccess(options)

        # Unregister Property
        await self.unregisterProperty(name)

        # Adapt the Topic:
        if not "topic" in options:
            # Raise an error, because we expect and dict or an string
            options.topic = getPropertyPath(self.identifier, name)

        if not "mode" in options:
            options.mode = ["publish", "subscrie"]

        if isinstance(options.topic, str) and not isPropertyPathCorrect(
                self.identifier, options.topic):
            # Adapt the name.
            options.topic = getPropertyPath(self.identifier, options.topic)
        else:
            if 'subscribe' in options.topic and not isPropertyPathCorrect(
                    self.identifier, options.topic.subscribe):
                options.topic.subscribe = getPropertyPath(
                    self.identifier, options.topic.subscribe)

            if 'publish' in options.topic and not isPropertyPathCorrect(
                    self.identifier, options.topic.publish):
                options.topic.publish = getPropertyPath(
                    self.identifier, options.topic.publish)

        observable = self._core.dataDistributor.register(observable, options)

        # Register the new Property.
        self._registeredProperties[name] = {
            'observable': observable,
            'options': options
        }

    async def registerEvent(self, name: str, emitter, options=None):
        """ Function to register an emitter as property. This will provide the emitter in the NoPE-Environment.

        Args:
            name (str): Name of the property, which should be used to forward the data
            emitter (nope-emitter): the property to use
            options (dict): options, to assign a different topic and options to select, whether the content of the property should be "published" and/or "subscribed"

        Returns:
            [type]: [description]
        """
        options = ensureDottedAccess(options)

        # Unregister Property
        await self.unregisterEvent(name)

        # Adapt the Topic:
        if not "topic" in options:
            # Raise an error, because we expect and dict or an string
            options.topic = getEmitterPath(self.identifier, name)
        elif isinstance(options.topic, str) and not isEmitterPathCorrect(self.identifier, options.topic):
            # Adapt the name.
            options.topic = getEmitterPath(self.identifier, options.topic)
        else:
            if 'subscribe' in options.topic and not isEmitterPathCorrect(
                    self.identifier, options.topic.subscribe):
                options.topic.subscribe = getEmitterPath(
                    self.identifier, options.topic.subscribe)

            if 'publish' in options.topic and not isEmitterPathCorrect(
                    self.identifier, options.topic.publish):
                options.topic.publish = getEmitterPath(
                    self.identifier, options.topic.publish)

        emitter = self._core.eventDistributor.register(emitter, options)

        # Register the new Property.
        self._registeredEvents[name] = {
            'emitter': emitter,
            'options': options
        }

    async def registerMethod(self, name: str, method, options=None):

        options = ensureDottedAccess(options)

        # Unregister the Function
        await self.unregisterMethod(name)

        # Adapt the Method ID
        if options.id is not None:
            if not isMethodPathCorrect(self.identifier, options.id):
                options.id = getMethodPath(self.identifier, options.id)
        else:
            options.id = getMethodPath(self.identifier, name)

        method = await self._core.rpcManager.registerService(method, options)

        # Register the new Function.
        self._registeredMethods[name] = {
            'method': method,
            'options': options
        }

    async def unregisterMethod(self, name: str):
        """ Unregister a method. (This is the case if the class provides dynamic classes)


        Args:
            name (str): Name of the function used during registering.
        """
        # Test if the Method is already registerd,
        # If so => unregister it first.
        if name in self._registeredMethods:
            await self._core.rpcManager.unregisterService(self._registeredMethods[name]['method'])

    async def unregisterProperty(self, name: str):
        """ Helper Function to unregister an Observable (a Property.)

        Args:
            name (str): Name of the Property used during registering.
        """
        # Test if the Property is already registerd,
        # If so => unregister it first.
        if name in self._registeredProperties:
            self._core.dataDistributor.unregister(
                self._registeredProperties[name]['observable'])

    async def unregisterEvent(self, name: str):
        """ Helper Function to unregister an Eventbased Property

        Args:
            name (str):  Name of the Property used during registering.
        """
        # Test if the Property is already registerd,
        # If so => unregister it first.
        if name in self._registeredProperties:
            self._core.eventDistributor.unregister(
                self._registeredProperties[name]['emitter'])

    async def listFunctions(self):
        return list(self._registeredMethods.values())

    async def listProperties(self):
        return list(self._registeredProperties.values())

    async def listEvents(self):
        return list(self._registeredEvents.values())

    async def init(self):
        # In self base Implementation, check if every requried property is set
        # correctly. If not => raise an error.
        self._logger = getNopeLogger(
            self.__class__.__name__ + '<' + self.identifier + '>')

        if self.type is None:
            raise Exception(
                'Please Provide a Name for the Module before initializing')

        if self.description is None:
            raise Exception(
                'Please Provide a Description for the Module before initializing')

        if self.author is None:
            raise Exception(
                'Please Provide an Author for the Module before initializing')

        if self.version is None:
            raise Exception(
                'Please Provide a Version for the Module before initializing')

        if self.identifier is None:
            raise Exception(
                'Please Provide an Identifier for the Module before initializing')

        # TODO: Test the Implementation with decorators.
        if self.decoratedItems:
            for item in self.decoratedItems:

                # Currently our decorators will add the information to all
                # classes...
                if item.owner != self.__class__:
                    # FIX
                    continue

                if item.type == "method":
                    await self.registerMethod(item.accessor, getattr(self, item.accessor), item.options)
                elif item.type == "property":
                    await self.registerProperty(item.accessor, getattr(self, item.accessor), item.options)
                elif item.type == "event":
                    await self.registerEvent(item.accessor, getattr(self, item.accessor), item.options)

    def exportProperty(self, name: str, observable, options):
        """ Function to export an Property. This will provide the Property (emits and receives events on change, and contains the current value.)
            in the NoPE-Environment.

        Args:
            name (str): Name of the property, which should be used to forward the data
            observable (nope-observable): the property to use
            options (dict-like): Options, used during sharing
        """
        # Create a Factory of the Element
        EXECUTOR.callParallel(self.registerProperty, name, observable, options)

    def exportMethod(self, name: str, func, options):
        """ Helper Function, to export method and provided it in the NoPE-Environment. Executes "registerMethod" in the background

        Args:
            name (str): Name of the property, which should be used as service name.
            func (callable): The method itself
            options (dict-like): Options, used during sharing
        """
        EXECUTOR.callParallel(self.registerMethod, name, func, options)

    def exportEvent(self, name: str, emitter, options):
        """ Function to export an Property as Event-Emitter (which emits en receives events). This will provide the Property in the NoPE-Environment.

        Args:
            name (str): Name of the property, which should be used to forward the data
            emitter (nope-emitter): the property to use
            options (dict-like): Options, used during sharing
        """
        # Create a Factory of the Element
        EXECUTOR.callParallel(self.registerProperty, name, emitter, options)

    async def dispose(self):
        """ Disposes the instance. This results in unregistering the properties, functions etc.
        """
        # Unregister all Methods and Functions
        for name in self._registeredMethods.keys():
            await self.unregisterMethod(name)

        # Remove all known Functions
        self._registeredMethods = dict()

        # Unregister all Properties.
        for name in self._registeredProperties.keys():
            await self.unregisterProperty(name)

        # Remove all known Properties.
        self._registeredProperties = dict()

        # Unregister all Properties.
        for name in self._registeredEvents.keys():
            await self.unregisterEvent(name)

        # Remove all known Properties.
        self._registeredEvents = dict()

    def getIdentifierOf(self, prop_event_or_method, type=None):
        """ Helper to extract the identifier for the given method, event or proterty
            or property.
        """

        # To Extract the name of the Property or the Function, we will iterate over
        # the registered properties and the regiestered functions. If the prop or the
        # function matches ==> return the name otherwise we throw an error.

        for data in self._registeredProperties.values():
            if data["observable"] == prop_event_or_method:
                # Extract the Topics, pipe and scope.
                _subTopic = data["options"]['topic'] if isinstance(
                    data["options"]['topic'], str) else data["options"]['topic'].get(
                    ['subscribe'], None)
                _pubTopic = data["options"]['topic'] if isinstance(
                    data["options"]['topic'], str) else data["options"]['topic'].get(
                    ['publish'], None)

                if type == "topic_to_subscribe":
                    if _subTopic is None:
                        raise Exception("No topic for subscribing defined")

                    return _subTopic

                elif type == "topic_to_publish":
                    if _pubTopic is None:
                        raise Exception("No topic for publishing defined")

                    return _pubTopic

                elif type == "both":
                    if not isinstance(data["options"]["topic"], str):
                        return data["options"]["topic"]

                    raise Exception(
                        "Prop uses the Same Element")

                else:
                    if isinstance(data["options"]["topic"], str):
                        return data["options"]["topic"]

                    raise Exception(
                        "Prop uses different name of Publishing an Subscribing")

        for data in self._registeredEvents.values():
            if data["emitter"] == prop_event_or_method:
                # Extract the Topics, pipe and scope.
                _subTopic = data["options"]['topic'] if isinstance(
                    data["options"]['topic'], str) else data["options"]['topic'].get(
                    ['subscribe'], None)
                _pubTopic = data["options"]['topic'] if isinstance(
                    data["options"]['topic'], str) else data["options"]['topic'].get(
                    ['publish'], None)

                if type == "topic_to_subscribe":
                    if _subTopic is None:
                        raise Exception("No topic for subscribing defined")

                    return _subTopic

                elif type == "topic_to_publish":
                    if _pubTopic is None:
                        raise Exception("No topic for publishing defined")

                    return _pubTopic

                elif type == "both":
                    if not isinstance(data["options"]["topic"], str):
                        return data["options"]["topic"]

                    raise Exception(
                        "Prop uses the Same Element")

                else:
                    if isinstance(data["options"]["topic"], str):
                        return data["options"]["topic"]

                    raise Exception(
                        "Prop uses different name of Publishing an Subscribing")

        for data in self._registeredMethods.values():
            if prop_event_or_method == data["func"]:
                return data["options"]["id"]

        raise Exception("Item not found. Is it registered?")

    def toDescription(self):
        ret = {
            'author': self.author,
            'description': self.description,
            'methods': self.methods,
            'events': self.events,
            'identifier': self.identifier,
            'properties': self.properties,
            'type': self.type,
            'version': self.version,
            'uiLinks': self.uiLinks
        }

        return ret



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from ...modules import BaseModule
from ...observable import NopeObservable


class HelloWorldModule(BaseModule):

    def __init__(self, core):
        super().__init__(core)

        # Define the Description:
        self.setAuthor("martin", "karkowski", "m.karkowski@zema.de")
        self.setVersion("1.0", "01.11.2022")
        self.description = "A Simple Hello World Module. It contains a Property and a Function."

        # Generate the Observable
        self.prop = NopeObservable()

    async def init(self):
        await super().init()

        await self.registerProperty('prop', self.prop, {
            'topic': 'prop',
            'mode': ['publish', 'subscribe']
        })

        await self.registerMethod('hello', self.hello, {
            "schema": {
                "type": "string",
                "name": "greetings_to",
                "description": "The person how receives the greeting"
            }
        })

        def _callback(data, rest):
            if (rest.sender != self.prop.id):
                self._logger.warn("External-Update. Data= " + str(data))

        self.prop.subscribe(_callback)

    async def hello(self, greetings_to: str):
        return 'hello ' + greetings_to + ' from ' + self.identifier

    async def dispose(self):
        await super().dispose()
        print(self.identifier, "as helloworld is getting removed")



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from .main import main_cli
from .run import generateNopeBackend, getDefaultParameters, run_cli, getArgs as getRunNopeBackendArgs
from .scan import create_config, list_packages, scan_cli
from .logo import renderLogo

__all__ = ["run_cli", "create_config", "list_packages",
           "scan_cli", "main_cli", "generateNopeBackend", "getDefaultParameters", "getRunNopeBackendArgs"]



from .stringMethods import replaceAll, camelToSnake, snakeToCamel, toVariableName
from typing import Literal

SPLITCHAR = '/'
SINGLE_LEVEL_WILDCARD = '+'
MULTI_LEVEL_WILDCARD = '#'


def pathToSnakeCase(path: str, splitchar=SPLITCHAR) -> str:
    # Join the splitchar with the adapted strings.
    return splitchar.join(
        map(
            lambda item: camelToSnake(item),
            path.split(splitchar)
        )
    )


def pathToCamelCase(path: str, splitchar=SPLITCHAR) -> str:
    # Join the splitchar with the adapted strings.
    return splitchar.join(
        map(
            lambda item: snakeToCamel(item),
            path.split(splitchar)
        )
    )


def convertPath(path: str) -> str:
    """ converts the path

    Args:
        path (str): The Path to adapt.

    Returns:
        str: The adapted Path
    """
    return replaceAll(path, ['.', '[', [']', '']], SPLITCHAR)


def _isNumber(item: str) -> bool:
    try:
        float(item)
        return True
    except BaseException:
        return False


def toPythonPath(path: str, style: Literal["dot", "bracket"] = "dot") -> str:
    """ converts the nope-path to a python path.

    Args:
        path (str): The Path to adapt.
        style (Literal["dot","bracket"], optional): The access-style (bracket or dotted.). Defaults to "dot".

    Returns:
        str: The adapted Path
    """

    ret = ""
    splitted = path.split(SPLITCHAR)

    for item in splitted:
        if len(ret) > 0:
            if _isNumber(ret):
                ret += f"[{item}]"
            elif style == "dot":
                ret += "." + item
            else:
                ret += f"[{item}]"
        else:
            ret = item
    return ret


def containsWildcards(str: str) -> bool:
    """ Determines, whether the given string contains a single level card or not.

    Args:
        str (str): String to check

    Returns:
        bool: Result
    """
    return SINGLE_LEVEL_WILDCARD in str or MULTI_LEVEL_WILDCARD in str


def getLeastCommonPathSegment(
        pathes, considerSingleLevel=False, considerMultiLevel=False):
    """ Returns the least common segmet of all pathes, included in the pathes array.

    Args:
        pathes (str[]): The Segments to compare.
        considerSingleLevel (bool): allows "singlelevel"-wildcards in the segments
        considerMultiLevel (bool): allows "multilevel"-wildcards in the segments

    Returns:
        (False | str): the least common segment of the pathes or False.
    """
    current_path = pathes.pop()
    while len(pathes) > 0:
        next = pathes.pop()
        current_path = _getLeastCommonPathSegment(
            current_path, next, considerSingleLevel=considerSingleLevel, considerMultiLevel=considerMultiLevel)

        # Only proceed, if there are elements included.
        if not current_path:
            # Return False
            return current_path

    # Return the least common segments.
    return current_path


def _getLeastCommonPathSegment(
        path01, path02, considerSingleLevel=False, considerMultiLevel=False):
    """_summary_

    Args:
        path01 (_type_): _description_
        path02 (_type_): _description_
        opts (_type_, optional): _description_. Defaults to dotted_dict({}).

    Returns:
        _type_: _description_
    """
    p1 = convertPath(path01).split(SPLITCHAR)
    p2 = convertPath(path02).split(SPLITCHAR)

    ret = []
    idx = 0

    l1 = len(p1)
    l2 = len(p2)

    _max = min(l1, l2)

    while idx < _max:

        if p1[idx] == p2[idx]:
            ret += [p1[idx]]
        elif considerSingleLevel:
            if p1[idx] == SINGLE_LEVEL_WILDCARD:
                ret += [p2[idx:]]
            elif p2[idx] == SINGLE_LEVEL_WILDCARD:
                ret += [p1[idx:]]
            else:
                break
        elif considerMultiLevel:
            if p1[idx] == MULTI_LEVEL_WILDCARD:
                ret += [p2[idx:]]
                break
            elif p2[idx] == MULTI_LEVEL_WILDCARD:
                ret += [p1[idx:]]
                break
            else:
                break
        else:
            break
        idx = idx + 1
    if len(ret):
        return SPLITCHAR.join(ret)
    return False


def patternIsValid(str: str) -> bool:
    """ Function to test if a pattern is valid

    Args:
        str (str): The pattern to test

    Returns:
        bool: The result.
    """

    if str == "":
        return True

    splitted = str.split(SPLITCHAR)
    last_index = len(splitted) - 1

    for idx, segment in enumerate(splitted):
        if segment:
            if segment == MULTI_LEVEL_WILDCARD:
                return idx == last_index
        else:
            return False

    return True


def varifyPath(path: str) -> str:
    """ Start to varify the Path.

    Args:
        path (str): The Path to adapt.

    Returns:
        str: The item to return
    """
    return SPLITCHAR.join(
        map(lambda item: toVariableName(item), path.split(SPLITCHAR)))



from . import baseServices, connectivityManager, core, instanceManager, nopeDispatcher, rpcManager
from .connectivityManager import NopeConnectivityManager
from .core import NopeCore
from .getDispatcher import getDispatcher
from .instanceManager import NopeInstanceManager
from .nopeDispatcher import NopeDispatcher
from .rpcManager import NopeRpcManager



from dataclasses import is_dataclass, asdict
from inspect import isfunction, getsource
from json import dumps as _dumps, loads as _loads, JSONEncoder, JSONDecoder
from typing import Any

FUNC_BEGIN = "__FUNCTION_BEGIN__"
FUNC_END = "__FUNCTION_END__"


class EnhancedJSONEncoder(JSONEncoder):
    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        if isfunction(o):
            code = getsource(o)

            if code.startswith("def"):

                code = code.replace("\n", "\\n")

                return f"{FUNC_BEGIN}{code}{FUNC_END}{o.__name__}"

            elif "lambda" in code:
                idx = code.index("lambda")
                return f"{FUNC_BEGIN}{code[idx:]}{FUNC_END}"
        return super().default(o)


class SimpleJSONEncoder(JSONEncoder):
    def default(self, o):
        if is_dataclass(o):
            return asdict(o)
        elif isfunction(o):
            return None
        return super().default(o)


class EnhancedJSONDecoder(JSONDecoder):
    def __init__(self, *args, **kwargs):
        JSONDecoder.__init__(
            self,
            object_hook=self.object_hook,
            *args,
            **kwargs)

    def object_hook(self, o):
        if isinstance(o, dict):

            to_adapt = dict()

            for key, value in o.items():
                if isinstance(value, str):
                    if value.startswith(FUNC_BEGIN) and FUNC_END in value:
                        code = value[len(FUNC_BEGIN):value.index(FUNC_END)]
                        if code.startswith("def"):
                            exec(code)
                            name = value[value.index(
                                FUNC_END) + len(FUNC_END):]
                            o[key] = eval(name)
                        else:
                            o[key] = eval(code)
                        continue
        return o


def dumps(o: Any, indent: int = 4, parse_functions=True, **kwargs) -> str:
    """ Helper function, to dump data to JSON and serialize
        python functions.


    Args:
        o (Any): The Object to start
        indent (int, optional): The indent. Defaults to 4.
        parse_functions (bool, optional): Flag to enable storing Python Functions. Defaults to True.

    Returns:
        str: The parsed object.
    """
    if parse_functions:
        kwargs.pop("cls", None)
        return _dumps(o, indent=indent, cls=EnhancedJSONEncoder, **kwargs)
    return _dumps(o, indent=indent, cls=SimpleJSONEncoder, **kwargs)


def loads(s: str, parse_functions=True, **kwargs) -> Any:
    if parse_functions:
        kwargs.pop("cls", None)
        return _loads(s, cls=EnhancedJSONDecoder, **kwargs)
    return _loads(s, **kwargs)



import math
import re
from typing import Union

from .dottedDict import DottedDict

pattern = re.compile(r'(?<!^)(?=[A-Z])')


def snakeToCamel(input: str) -> str:
    """Converts the given snake-case to camel-case
    """
    # split underscore using split
    temp = input.split('_')

    # joining result
    res = temp[0] + ''.join(ele.title() for ele in temp[1:])

    # Return the Result
    return res


def camelToSnake(input: str) -> str:
    """Converts the given camel case to snake-case
    """
    return pattern.sub('_', input).lower()


def replaceAll(org_str: str, value, replacement: str) -> str:
    """ Replaces all Chars in a String

    Args:
        org_str (str): base string
        value (str): the value which should be replaced, Or an array containing the value and replacer
        replacement (str): the value which is used as replacement

    Returns:
        str: The adapted Place
    """

    ret: str = org_str

    if isinstance(value, str) and (replacement is not None):
        ret = ret.replace(value, replacement)
    else:
        for pair in value:
            if isinstance(pair, str) and (replacement is not None):
                ret = ret.replace(pair, replacement)
            else:
                ret = ret.replace(pair[0], pair[1])

    return ret


def padString(num: Union[float, int], size: int, max_length=False):
    """ Function to Pad a String.

    Args:
        num (Union[float,int]): number
        size (int): Max pad-size or max number
        max_length (bool, optional): Flat must be set to true, if the size is the max value. Defaults to False.

    Returns:
        str: the Number as String
    """
    size = size
    if isinstance(max_length, bool) and max_length:
        size = math.ceil(math.log10(size))
    s = str(num)
    while len(s) < size:
        s = '0' + s
    return s


def insert(str: str, index: int, content: str) -> str:
    """ Inserts a String in the String
    """
    if index > 0:
        return str[0, index] + content + str[index, len(str)]
    else:
        return content + str


def toCamelCase(str: str, char='_') -> str:
    def callback_0(word, index):
        return word.to_lower_case() if index == 0 else word.to_upper_case()

    return replaceAll(str, char, ' ').replace(re.compile(
        '(?:^\\w|[A-Z]|\\b\\w)'), callback_0).replace(
        re.compile('\\s+'), '')


def toSnakeCase(str: str) -> str:
    """ Helper to convert the string to the snake-case

    Args:
        str (str): The string to convert

    Returns:
        str: The converted string
    """
    if str == str.upper():
        return str

    ret = ''.join(['_' + i.lower() if i.isupper()
                   else i for i in str]).lstrip('_')

    if ret.startswith("_"):
        ret[1:]

    return ret


def limitString(str: str, length: int,
                limit_chars: str = '...') -> DottedDict:
    """ Helper to limit the string to a specific length. the rest is reduced by the limitChars

    Args:
        str (str): The string to work with
        length (int): The max length including the limitChars
        limit_chars (str, optional): The chars which should be used to express limiting. Defaults to '...'.

    Returns:
        dotted_dict: dict containing the following keys 'isLimited', 'original', 'limited'
    """
    if len(str) > length:
        return DottedDict({
            'isLimited': True,
            'original': str,
            'limited': str[0, length - len(limit_chars)] + limit_chars
        })
    else:
        return DottedDict({
            'isLimited': False,
            'original': str,
            'limited': str
        })


def insertNewLines(str: str, max_length: int = 100) -> str:
    """ Helper to insert new lines after a given amount of time.

    Args:
        str (str): The String to work with
        max_length (int, optional): The Max-Lenght of the string, after which a new-line symbole has to be inserted. Defaults to 100.

    Returns:
        str: The adapted string.
    """

    splitted = str.split(' ')
    ret = []
    length = 0
    for word in splitted:

        length = length + len(word) + 1

        ret.append(word)

        if length > max_length:
            ret.append('\\n')
            length = 0

    return ret


TO_REPLACE = {
    'ä': 'ae',
    'ü': 'ue',
    'ö': 'oe',
    'ß': 'ss'
}


def toVariableName(s: str) -> str:
    """ Helper to convert the Variable to a valid varialbe name.

    Example:
        >>> toVariableName('32v2 g #Gmw845h$W b53wi ')
        '_32v2_g__Gmw845h_W_b53wi_'

    Args:
        s (str): The String to convert

    Returns:
        str: The string as valid variable name.
    """
    ret = re.sub('\\W|^(?=\\d)', '_', s)
    for k, v in TO_REPLACE.items():
        ret = ret.replace(k, v)

    return ret



""" An example how to modify the Behavior of multiple elements using a Plugin.

    In This case the Plugin allows the implementation of an ackknowledgement
    message if required.
"""

import asyncio

from nope.eventEmitter import NopeEventEmitter
from nope.helpers import generateId, EXECUTOR, ensureDottedAccess, formatException, isAsyncFunction
from nope.plugins import plugin

_DEFAULT_RESULT = object()


@plugin([
    "nope.dispatcher.rpcManager",
    "nope.dispatcher.connectivityManager"
],
    name="rpcCallbacks")
def extend(rpcMod, conManagerMod):
    "Extends the Bridge and adds the functionality of ack knowlededing messages."

    class NopeRpcManager(rpcMod.NopeRpcManager):
        def __init__(self, *args, **kwargs):
            rpcMod.NopeRpcManager.__init__(self, *args, **kwargs)
            self.defaultKeepAlive = kwargs.get(
                "defaultKeepAlive", 60 * 60 * 1000)

        async def _handleExternalRequest(self, data, func=None):
            try:
                if not callable(func):
                    if data.functionId not in self._registeredServices:
                        return
                    func = self._registeredServices[data.functionId].get(
                        "func", None)

                if self._logger:
                    self._logger.debug(
                        f'Dispatcher "{self.id}" received request: "{data.functionId}" -> task: "{data.taskId}"')

                if callable(func):
                    # Now we check, if we have to perform test, whether
                    # we are allowed to execute the task:
                    if data.get("target", self.id) != self.id:
                        return

                    # Define a list containing callbacks:
                    cbs = []

                    observer = None

                    def on_cancel(reason, *args):
                        nonlocal observer
                        if reason.taskId == data.taskId:

                            for cb in cbs:
                                cb(reason)

                            observer.unsubscribe()
                            observer = None

                    observer = self.onCancelTask.subscribe(on_cancel)

                    # extract the arguments etc.
                    # create an empty list and fill it afterwards
                    args = [None] * (len(data.params) + len(data.callbacks))

                    for item in data.params:
                        args[item.idx] = item.data

                    for optionsOfCallback in data.callbacks:

                        async def callback(*args, id=optionsOfCallback.id, opts=optionsOfCallback, **kwargs):
                            nonlocal cbs

                            servicePromise = self.performCall(id, args, opts)

                            def cancelCallback(reason):
                                # The Main Task has been canceled =>
                                # We are allowed to canel the Subtask as well.
                                if servicePromise is not None and getattr(
                                        servicePromise, 'cancelCallback', False):
                                    servicePromise.cancelCallback(reason)

                            idx = len(cbs)
                            cbs.append(cancelCallback)

                            # Await the Result. If an Task is canceled => The
                            # Error is Thrown.
                            result = await servicePromise

                            # Remove the Index
                            cbs.pop(idx)

                            return result

                        args[optionsOfCallback.idx] = callback

                    # Perform the Task it self.
                    _result = _DEFAULT_RESULT

                    if not func.isAsync and not self.__warned and self._logger:
                        self._logger.warn(
                            "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                        self._logger.warn(
                            "Offloading function to a separate thread.")
                        # We only want to warn the user once.
                        self.__warned = True

                    resultPromise = func(*args)

                    try:
                        if resultPromise is not None and getattr(
                                resultPromise, 'cancelCallback', False):
                            def _cancel_main(reason):
                                resultPromise.cancelCallback(reason)

                            cbs.append(_cancel_main)

                    except Exception as error:
                        # The Cancel Function isn't available in
                        # the provided promise.
                        pass

                    self._runningExternalRequestedTasks[data.taskId] = data.requestedBy

                    # Wait for the Result to finish.
                    _result = await resultPromise

                    # Define the Result message
                    result = {
                        'result': _result if _result is not _DEFAULT_RESULT else None,
                        'taskId': data.taskId,
                        'type': 'response'
                    }

                    if self._logger:
                        self._logger.debug(
                            'Internally executed requested Function for Task: ' + str(data['taskId']) + " - Function \"" +
                            data['functionId'] + '\". Sending result on ' + data['resultSink'])

                    # Use the communicator to publish the result.
                    await self._communicator.emit("rpcResponse", result)

            except Exception as error:

                if self._logger:
                    self._logger.error(
                        f'Dispatcher "{self.id}" failed with request: "{data.taskId}"')
                    self._logger.error(formatException(error))
                else:
                    print(formatException(error))

                self._runningExternalRequestedTasks.pop(data.requestedBy, None)

                result = {
                    'error': {
                        'error': str(error),
                        'msg': str(error)
                    },
                    'taskId': data.taskId,
                    'type': 'response'
                }

                # Use the communicator to publish the result.
                await self._communicator.emit("rpcResponse", result)

        async def _performCall(self, serviceName, params, options=None):
            optionsToUse = ensureDottedAccess({
                "resultSink": self._getServiceName(serviceName, "response"),
                "waitForResult": True,
                "callbackOptions": [],
                "timeToLifeAfterCall": self.defaultKeepAlive,
                "calledOnce": [],
            })
            optionsToUse.update(ensureDottedAccess(options))

            taskId = generateId()

            _registeredCallbacks = []

            # Create a Future of the Loop.
            future = EXECUTOR.generatePromise(taskId=taskId)

            def clear():
                if taskId in self._runningInternalRequestedTasks:

                    if self._logger:
                        self._logger.debug('Clearing Callbacks from ' + taskId)

                    for cb in _registeredCallbacks:
                        EXECUTOR.callParallel(self.unregisterService, cb)

                    task = self._runningInternalRequestedTasks[taskId]

                    if task.timeout is not None:
                        task.timeout.cancel()

                    self._runningInternalRequestedTasks.pop(taskId)

            try:
                tastRequest = ensureDottedAccess({
                    "future": future,
                    "clear": clear,
                    'serviceName': serviceName,
                    'timeout': None,
                })

                # Store the Future as Task.
                self._runningInternalRequestedTasks[taskId] = tastRequest

                # Define the packet to send:
                packet = {
                    'functionId': serviceName,
                    'params': [],
                    'taskId': taskId,
                    'resultSink': optionsToUse['resultSink'],
                    'requestedBy': self._id,
                    'target': None,
                    'callbacks': []
                }

                callbackOptions = {
                    item.idx: item for item in optionsToUse.callbackOptions}

                # Iterate over all Parameters and
                # Determin Callbacks. Based on the Parameter-
                # Type assign it either to packet.params (
                # for parsable Parameters) and packet.callbacks
                # (for callback Parameters)
                for idx, contentOfParameter in enumerate(params):
                    if not callable(contentOfParameter):
                        # The parameter isnt callable:
                        packet['params'].append({
                            'idx': idx,
                            'data': contentOfParameter
                        })
                    else:

                        if not self.__warned and not isAsyncFunction(
                                contentOfParameter):
                            if self._logger:
                                self._logger.warn(
                                    "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                                self._logger.warn(
                                    f'The service of parameter {idx} is synchronous!')
                            else:
                                print(
                                    "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                                print(
                                    f'The service of parameter {idx} is synchronous!')
                            # We only want to warn the user once.
                            self.__warned = True

                        _functionAsAsync = EXECUTOR.wrapFuncIfRequired(
                            contentOfParameter)

                        _timeToLifeAfterCall = callbackOptions.get(
                            idx, {"timeToLifeAfterCall": optionsToUse.timeToLifeAfterCall})["timeToLifeAfterCall"]
                        _calledOnce = callbackOptions.get(
                            idx, {"calledOnce": idx in optionsToUse.calledOnce})["calledOnce"]

                        # The Parameter is a Callback => store a
                        # Description of the Callback and register
                        # the callback inside of the Dispatcher

                        _id = ""
                        _timeout: asyncio.Task = None

                        def removeCallback():
                            nonlocal _id
                            EXECUTOR.callParallel(self.unregisterService, _id)

                        async def cbWithTimeout(*args, f=_functionAsAsync, t=_timeToLifeAfterCall, **kwargs):
                            nonlocal _timeout

                            if _timeout:
                                _timeout.cancel()

                            if t > 0:
                                _timeout = EXECUTOR.setTimeout(
                                    removeCallback, t)

                            return await f(*args, **kwargs)

                        async def cbOnce(*args, f=_functionAsAsync, **kwargs):

                            res = await f(*args, **kwargs)
                            await self.unregisterService(_id)
                            return res

                        if _timeToLifeAfterCall > 0:
                            _timeout = EXECUTOR.setTimeout(
                                removeCallback, _timeToLifeAfterCall)

                        _func = await self.registerService(cbOnce if _calledOnce else cbWithTimeout, {
                            "schema": {},
                            "id": generateId(pre_string="callback")
                        })

                        # Assign the used id.
                        _id = _func.id

                        _registeredCallbacks.append(_func.id)

                        # Register the Callback itself in the params object.
                        packet['callbacks'].append(
                            # Convert the Options to
                            # Camel Case
                            {
                                "id": _id,
                                "idx": idx
                            }
                        )

                if not self.serviceExists(serviceName):
                    error = Exception(
                        f'No Service Provider known for "{serviceName}"')
                    if self._logger:
                        self._logger.error(
                            f'No Service Provider known for "{serviceName}"')
                        self._logger.error(formatException(error))

                    raise error

                if self.options.forceUsingSelectors or self.services.amountOf.get(
                        serviceName, 0) > 1:

                    optionsForSelector = ensureDottedAccess({
                        "rpcManager": self,
                        "serviceName": serviceName
                    })

                    if isinstance(optionsToUse.target, str):
                        tastRequest.target = optionsToUse.target
                    elif callable(optionsToUse.selector):
                        tastRequest.target = await options.selector(optionsForSelector)
                    elif isinstance(optionsToUse.selector, str):
                        tastRequest.target = await self._defaultSelector(optionsForSelector)

                else:
                    tastRequest.target = list(
                        self.services.keyMappingreverse[serviceName])[0]

                packet["target"] = tastRequest.target

                await self._communicator.emit("rpcRequest", packet)

                if self._logger:
                    self._logger.debug(
                        f'Dispatcher "{self._id}" putting task "{taskId}" on: "{self._getServiceName(serviceName, "request")}"')

                if optionsToUse.get("timeout", 0) > 0:
                    async def onTimeout():
                        await self.cancelTask(
                            taskId,
                            TimeoutError(
                                f"TIMEOUT. The Service allowed execution time of {optionsToUse.timeout} [ms] has been excided")
                        )

                    # Create our timeout and store it.
                    tastRequest.timeout = EXECUTOR.setTimeout(
                        onTimeout, optionsToUse.timeout)

            except Exception as err:
                if self._logger:
                    self._logger.error('Something went wrong on calling')
                    self._logger.exception(formatException(err))
                else:
                    print(formatException(err))

                # Call the Clear Function
                clear()

                # Throw an error.
                future.set_exception(err)

            # Define a Function, which could be used for cancelation.
            def _cancelTask(reason):
                EXECUTOR.callParallel(self.cancelTask, taskId, reason)

            future.cancelCallback = _cancelTask

            if not optionsToUse.waitForResult:
                EXECUTOR.callParallel(future)
                return future

            return await future

    class NopeConnectivityManager(conManagerMod.NopeConnectivityManager):
        def _info(self):
            ret = conManagerMod.NopeConnectivityManager._info(self)
            ret.plugins.append("rpcCallbacks")
            return ret

    return NopeRpcManager, NopeConnectivityManager



import time

from nope.helpers import Emitter, generateId
from nope.observable import NopeObservable

PROFILE = False
OPEN_REQUESTS = {}


def profile_task(mode: str, data):
    try:
        if mode == "new":
            OPEN_REQUESTS[data["taskId"]] = time.process_time()
        elif mode == "done":
            if data["taskId"] in OPEN_REQUESTS:
                start = OPEN_REQUESTS.pop(data["taskId"])
                end = time.process_time()
                delta = end - start

                print(
                    "Bearbeitung hat",
                    round(
                        delta * 1000,
                        2),
                    "[ms] gedauert")
    except BaseException as e:
        print(f"Failed with mdoe= {mode}, data={data} cause: '{str(e)}'")


class EventCommunicationInterface:

    def __init__(self, emitter, logger=None, receivesOwnMessages=True):
        self._emitter = emitter if emitter is not None else Emitter()
        self._logger = logger
        self.receivesOwnMessages = receivesOwnMessages
        self.connected = NopeObservable()
        self.connected.setContent(True)
        self.id = generateId()

    async def on(self, eventName: str, cb):

        if PROFILE:
            if eventName == "rpcRequest":
                def req(data, *args):
                    profile_task("new", data)
                    cb(data)

                self._emitter.on(eventName, req)
            elif eventName == "rpcResponse":
                def res(data, *args):
                    profile_task("done", data)
                    cb(data)

                self._emitter.on(eventName, res)
            else:
                self._emitter.on(eventName, cb)
        else:
            self._emitter.on(eventName, cb)

        if eventName != 'statusChanged' and self._logger:
            def loggingCallback(*args):
                self._logger.debug('received', "'" + eventName + "'", *args)

            self._emitter.on(eventName, loggingCallback)

    async def emit(self, eventName: str, data):

        if PROFILE:
            if eventName == "rpcRequest":
                profile_task("new", data)
            if eventName == "rpcResponse":
                profile_task("done", data)

        self._emitter.emit(eventName, data)

    async def dispose(self):
        self._emitter.close()

    def detailListeners(self, t, listeners):
        raise Exception('Method not implemented.')



import asyncio
from asyncio import sleep

import pytest

from nope.plugins import install

"""
This is an Example how to extend the behavior of any class inside of Nope.

Therefore we want to load our sample plugin: "hello".
Therefore we know
"""


# install(nope, "nope.plugins.hello", plugin_dest="nope.dispatcher.rpcManager")


@pytest.fixture
def event_loop():
    loop = asyncio.new_event_loop()
    nope.EXECUTOR.assignLoop(loop)
    yield loop
    loop.close()


async def test_bridge_plugin():
    import nope
    nope, _, __ = install(nope, "nope.plugins.rpc_with_callbacks")

    dispatcher = nope.getDispatcher({
        "communicator": await nope.getLayer("event"),
        "logger": "info",
    })

    # To make our Plugin work -> we have to manually assign the id
    # dispatcher.communicator.ackReplyId = dispatcher.id
    await dispatcher.ready.waitFor()

    async def funcWithCallback(p, cb1, cb2):
        return await cb1(p, cb2)

    await dispatcher.rpcManager.registerService(funcWithCallback, {
        "id": "funcWithCallback"
    })

    async def cb(p, cb2):
        print("inside of callback")
        await cb2()

    await dispatcher.rpcManager.performCall("funcWithCallback", ["jeah", cb, lambda *args: print("Here")], {
        "calledOnce": [1]
    })


if __name__ == "__main__":
    import nope
    nope.EXECUTOR.loop.run_until_complete(test_bridge_plugin())



""" An example how to modify the Behavior of multiple elements using a Plugin.

    In This case the Plugin allows the implementation of an ackknowledgement
    message if required.
"""

from nope.eventEmitter import NopeEventEmitter
from nope.helpers import generateId, EXECUTOR, ensureDottedAccess, formatException
from nope.plugins import plugin


@plugin([
    "nope.communication.bridge",
    "nope.dispatcher.connectivityManager"
],
    name="ackMessages")
def extend(bridgeMod, conManagerMod):
    "Extends the Bridge and adds the functionality of ack knowlededing messages."

    class Bridge(bridgeMod.Bridge):
        def __init__(self, *args, **kwargs):
            bridgeMod.Bridge.__init__(self, *args, **kwargs)

            # Define
            self.defaultTargets = kwargs.get("defaultTargets", list())
            self.ackReplyId = kwargs.get("ackReplyId", None)
            self.onTransportError = NopeEventEmitter()
            self._onMessageReceived = NopeEventEmitter()

            # Dict containing the Messages, that have been send as key and containign the
            # still open respones (dispatcher ids) as set as value
            self._openMessages = {}

            # Make shure to forward ackknowledge messages.
            EXECUTOR.callParallel(
                self.on,
                "ackMessage",
                lambda msg: self._onMessageReceived.emit(msg))

            # Subscribe to Errors:
            def onTransportError(err, *args):
                if self._logger:
                    self._logger.error(
                        "Failed to receive an acknowledge message!")
                    self._logger.error(formatException(err))
                else:
                    print(formatException(err))

            self.onTransportError.subscribe(onTransportError)

        async def emit(self, eventName, data, target=None, timeout=0, **kwargs):
            promise = None

            if eventName != "ackMessage" and self.ackReplyId is not None:
                # Now lets define the Target:
                if target is None or target == True:
                    if self.defaultTargets:
                        target = set(self.defaultTargets)
                    elif "target" in data:
                        target = set([data["target"]])
                    else:
                        target = set()
                else:
                    if isinstance(target, str):
                        target = set([target])
                    elif target == False:
                        target = set()
                    else:
                        target = set(target)

                # Only if we expect a target,
                # we will wait for the message.
                if len(target) > 0:
                    messageId = generateId()
                    data["messageId"] = messageId

                    def callback(msg, *args):
                        nonlocal messageId

                        if msg.get(
                                "messageId", False) == messageId and messageId in self._openMessages:

                            data = self._openMessages.get(
                                messageId, False)

                            # Mark the Dispatcher as ready:
                            data["received"].add(msg["dispatcherId"])

                            if len(data["target"] - data["received"]) == 0:
                                # Remove the Message, becaus it is finished
                                self._openMessages.pop(messageId)

                                return True

                        return False

                    promise = self._onMessageReceived.waitFor(
                        callback, options={"timeout": timeout})

                    # Store the Message
                    self._openMessages[messageId] = {
                        "received": set(),
                        "target": target,
                        "promise": promise
                    }

            res = await bridgeMod.Bridge.emit(self, eventName, data, **kwargs)

            if promise:
                await promise

            return res

        async def on(self, eventName, cb):
            # In here we will define a custom callback,
            # which will send an "ackMessage" after
            # performing the original callback.

            if eventName == "ackMessage":
                return await bridgeMod.Bridge.on(self, eventName, cb)

            def callback(msg):
                cb(msg)

                if "messageId" in msg and self.ackReplyId is not None:
                    EXECUTOR.callParallel(
                        self.emit,
                        "ackMessage",
                        ensureDottedAccess({
                            "messageId": msg["messageId"],
                            "dispatcherId": self.ackReplyId
                        })
                    )

            return await bridgeMod.Bridge.on(self, eventName, callback)

    class NopeConnectivityManager(conManagerMod.NopeConnectivityManager):
        def __init__(self, *args, **kwargs):
            conManagerMod.NopeConnectivityManager.__init__(
                self, *args, **kwargs)

            self._communicator.ackReplyId = self._id
            self.forceAckMessage = True

            # To extend
            def cb(dispatchers: list, *args):
                if self.forceAckMessage:
                    # Get all Dispatchers running the Plugin.
                    dispatchersWithPlugin = list(
                        filter(
                            lambda id: "ackMessages" in self.dispatchers.originalData[id].plugins,
                            dispatchers))
                    self._communicator.defaultTargets = dispatchersWithPlugin

            self.dispatchers.data.subscribe(cb)

        def _info(self):
            ret = conManagerMod.NopeConnectivityManager._info(self)
            ret.plugins.append("ackMessages")
            return ret

    return Bridge, NopeConnectivityManager



#!\\usr\\bin\\env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import json
import os

from nope.helpers import dynamicImport, formatException, dumps, loads
from nope.logger import getNopeLogger

from .run import getDefaultParameters

LOGGER = getNopeLogger("config-scanner")


def list_packages(path: str, absolute = True):
    """ Function to scan for packages. Therefore a given has been provided

    Args:
        path (str): Path of the folder to scan.
        absolute (bool, optional): Store the folders absolute. Defaults to True.

    Returns:
        _type_: _description_
    """

    packages = []

    for root, directories, files in os.walk(path):
        if "/venv/" in root:
            continue

        for name in files:
            if name == "__init__.py":
                # Determine the Path to the File.
                pathToFile = os.path.join(root, name)

                if absolute:
                    pathToFile = os.path.abspath(os.path.join(root, name))
                else:
                    pathToFile = os.path.relpath(pathToFile)

                LOGGER.info(f"Checking file: '{pathToFile}'")

                try:
                    # Dynamically load the File.
                    module = dynamicImport(
                        name, pathToFile)

                    # Add the Path:
                    description = module.DESCRIPTION.__dict__
                    description["path"] = pathToFile

                    description = loads(
                        dumps(
                            description,
                            parse_functions=False))

                    # Add it to the packages
                    packages.append(description)
                except ModuleNotFoundError as E:
                    LOGGER.error(f"Failed to load file. '{pathToFile}'")
                    LOGGER.error(formatException(E))
                    pass
                except ImportError as e:
                    if e.msg == "attempted relative import with no known parent package":
                        pass
                    else:
                        pass
                        LOGGER.error(f"Failed to load file. '{pathToFile}'")
                        LOGGER.error(formatException(e))
                except AttributeError as E:                    
                    LOGGER.error(f"Failed to load file. '{pathToFile}'")
                    LOGGER.error(formatException(E))
                    pass
                except Exception as e:
                    LOGGER.error(f"Failed to load '{pathToFile}'")
                    LOGGER.error(formatException(e))
                    pass

    return packages


def create_config(path_to_scan: str, dest_dir: str, name_of_file: str, absolute = False):
    """ Create a configuration """

    LOGGER.info(f"Starting to find files in '{path_to_scan}'")
    packages = list_packages(path_to_scan, absolute)

    try:
        os.makedirs(dest_dir)
    except OSError as e:
        pass  # already exists

    # Define the File
    path_to_config_file = os.path.join(dest_dir, name_of_file)

    for package in packages:
        for cl in package["providedClasses"]:
            cl.pop("createInstance", None)

        for func in package["providedServices"]:
            func.pop("function", None)

    config = {
        "packages": packages,
        "config": getDefaultParameters(),
        "connections": []
    }

    with open(path_to_config_file, "w") as file:
        file.write(json.dumps(config, indent=4))


def scan_cli(add_mode=True):
    import argparse
    parser = argparse.ArgumentParser(description='Module-Scanner')

    if add_mode:
        parser.add_argument('mode', type=str, default="scan",
                            help='option, used to scan for modules')

    parser.add_argument('--input', type=str, default="./", dest='input',
                        help='folder to scan for modules')
    parser.add_argument('--name', type=str, default="settings.json", dest='name',
                        help='name of the configuration file')
    parser.add_argument('--output', type=str, default="config", dest='output',
                        help='output directory for the configuration file')
    parser.add_argument('-r', '--relative', dest='relative', action='store_true',
                        help='Stores relativ pathes.')

    args = parser.parse_args()

    # Create the config
    create_config(
        os.path.dirname(os.path.join(os.getcwd(), args.input)),
        os.path.join(os.getcwd(), args.output),
        args.name,
        not args.relative
    )


if __name__ == "__main__":
    scan_cli(False)



from .DecoratedHelloWorld import DecoratedHelloWorldModule
from .HelloWorld import HelloWorldModule
from ...types import ProvidedClass, NopePackage, DefaultInstance, ProvidedService, ServiceOptions


async def _createHellowordModule(dispatcher, *args):
    """ Helper Function to Create the instance. """
    return HelloWorldModule(dispatcher)


async def helloWorldService(msg: str):
    return "Hello " + msg + "!"


DESCRIPTION = NopePackage("helloWorld",
                          defaultInstances=[
                              DefaultInstance("instance01", "HelloWorldModule")
                          ],
                          providedClasses=[
                              ProvidedClass(
                                  "HelloWorldModule", _createHellowordModule)
                          ],
                          providedServices=[
                              ProvidedService(
                                  helloWorldService,
                                  ServiceOptions(
                                      id="helloWorldService",
                                      schema={}
                                  )
                              )
                          ])



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import asyncio
import json

from nope.communication.bridge import Bridge
from nope.dispatcher.connectivityManager import NopeConnectivityManager
from nope.dispatcher.rpcManager import NopeRpcManager
from nope.helpers import SPLITCHAR, ensureDottedAccess, generateId, isIterable, EXECUTOR, waitFor, varifyPath, formatException
from nope.logger import defineNopeLogger
from nope.merging import DictBasedMergeData
from nope.modules import NopeGenericModule
from nope.observable import NopeObservable


class NopeInstanceManager:

    def __init__(self, options, _defaultSelector, _id=None,
                 _connectivityManager=None, _rpcManager=None, _core=None):

        self.options = options
        self._defaultSelector = _defaultSelector
        self._id = _id
        self._connectivityManager = _connectivityManager
        self._rpcManager = _rpcManager
        self._core = _core
        self._communicator: Bridge = options.communicator
        self.__disposed = False

        if _id is None:
            self._id = generateId()
        if _connectivityManager is None:
            self._connectivityManager = NopeConnectivityManager(
                options, id=self._id)
        if _rpcManager is None:
            self._rpcManager = NopeRpcManager(
                options,
                self._defaultSelector,
                id=self._id,
                connectivityManager=self._connectivityManager
            )

        self._logger = defineNopeLogger(
            options.logger, 'core.instance-manager')

        # Flag to indicate, that the system is ready.
        self.ready = NopeObservable()
        self.ready.setContent(False)

        self._mappingOfRemoteDispatchersAndGenerators = dict()
        # Overview of the available Constructors in the network.
        self.constructors = DictBasedMergeData(
            self._mappingOfRemoteDispatchersAndGenerators, '+', '+')
        self._mappingOfRemoteDispatchersAndInstances = dict()

        # Overview of the available instances in the network.
        #   - OriginalKey = DispatcherID (string);
        #   - OriginalValue = Available Instance Messages (IAvailableInstancesMsg);
        #   - ExtractedKey = The name of the Instance (string);
        #   - ExtractedValue = instance-description (INopeModuleDescription);
        self.instances = DictBasedMergeData(
            self._mappingOfRemoteDispatchersAndInstances, 'instances/+', 'instances/+/identifier')

        self._internalWrapperGenerators = dict()
        self._registeredConstructors = dict()
        self._instances = dict()
        self._externalInstances = dict()
        self._internalInstances = set()
        self._initializingInstance = dict()
        self._externalInstancesNames = set()

        # Contains the identifiers of the instances, which are hosted in the
        # provided dispatcher.
        self.internalInstances = NopeObservable()
        self.internalInstances.setContent([])

        self.constructorServices = NopeObservable()
        self.constructorServices.setContent([])

        _ctorStart = f'nope{SPLITCHAR}core{SPLITCHAR}constructor{SPLITCHAR}'

        def _extractGenerators(*args):

            constructorServices = set()

            self._mappingOfRemoteDispatchersAndGenerators.clear()
            for dispatcher, services in self._rpcManager.services.originalData.items():
                def _filterMatchingServices(svc):
                    if "id" in svc and svc.id.startswith(_ctorStart):
                        constructorServices.add(svc["id"])
                        return True
                    return False

                generators = list(
                    map(
                        lambda item: item.id[len(_ctorStart):],
                        filter(
                            _filterMatchingServices,
                            services.services
                        )
                    )
                )

                if len(generators):
                    self._mappingOfRemoteDispatchersAndGenerators[dispatcher] = generators

            self.constructorServices.setContent(list(constructorServices))
            self.constructors.update()

        # Subscribe to changes.
        self._rpcManager.services.data.subscribe(_extractGenerators)

        if self._logger:
            self._logger.info('core.instance-manager online')

        self.reset()
        EXECUTOR.callParallel(self._init)

    async def _sendAvailableInstances(self):
        # Update the Instances provided by this module.
        await self._communicator.emit(
            "instancesChanged",
            {
                "dispatcher": self._id,
                # We will send the descriptions.
                # Generate the Module Description for every identifier:
                "instances": list(
                    map(lambda item: self._instances[item]["instance"].toDescription(), self._internalInstances))
            }
        )

        # Update the Instances
        self.internalInstances.setContent(list(self._internalInstances))

    async def _init(self):

        await self._communicator.connected.waitFor()
        await self._connectivityManager.ready.waitFor()
        await self._rpcManager.ready.waitFor()

        async def _generateWrapper(dispather, description):
            mod = NopeGenericModule(
                dispather,
                # self._generateEmitter,
                # self._generateObservable
            )
            await mod.fromDescription(description, "overwrite")
            return mod

        self.registerInternalWrapperGenerator(
            "*",
            _generateWrapper
        )

        def _onDispatchersChanged(changes, *args):
            """ Callback which will handle new and offline Dispatchers.
            """

            if len(changes.added):
                # If there are dispatchers online,
                # We will emit our available services.
                EXECUTOR.callParallel(self._sendAvailableInstances)

            if len(changes.removed):
                # Remove the dispatchers.
                for removedId in changes.removed:
                    self.removeDispatcher(removedId)

        # We will use our status-manager to listen to changes.
        self._connectivityManager.dispatchers.onChange.subscribe(
            _onDispatchersChanged)

        # Make shure we are emitting the instances provided.
        await self._communicator.on("bonjour", lambda *args: EXECUTOR.callParallel(self._sendAvailableInstances))

        def _onInstancesChanged(message, *args):
            """ Callback which will be called if the commincator receives a Message
                that some instances has been changed.

            Args:
                message: The Message from the System.
            """
            # Store the instance.
            self._mappingOfRemoteDispatchersAndInstances[message.dispatcher] = message

            # Update the Mapping
            self.instances.update()

            if self._logger:
                self._logger.debug(
                    'Remote Dispatcher "' + str(message.dispatcher) + '" updated its available instances')

        # Listen to the Changes.
        await self._communicator.on("instancesChanged", _onInstancesChanged)

        if self._logger:
            self._logger.debug("core.instance-manager " +
                               str(self._id) + " initialized")

        self.ready.setContent(True)

    def getServiceName(self, name: str, type: str) -> str:
        """ Helper to get the corresponding Service name

        Args:
            name (str): Name of the Service
            _type (str): The desired type of the requested service name. Could be "dispose" or "constructor"

        Returns:
            str: The Adapted Name
        """
        if type == "constructor":
            return f"nope{SPLITCHAR}core{SPLITCHAR}constructor{SPLITCHAR}{name}"
        elif type == "dispose":
            return f"nope{SPLITCHAR}core{SPLITCHAR}destructor{SPLITCHAR}{name}"
        else:
            raise Exception("The given type is not correct.")

    def _getInstanceInfo(self, identifier: str):
        """ Function, that will extract the information of the instance and the providing dispatcher.

        Args:
            identifier (str): The identifier of instance
        """
        # First check if the instance exists.
        if not self.instanceExists(identifier, False):
            return None

        ret = ensureDottedAccess({})

        # First we check if we are taking care of an internal instance, if so
        # we will use this instance to enrich the description, otherwise, we
        # will look in the external instances.
        if identifier in self._instances:
            ret.description = self._instances[identifier].instance.toDescription(
            )
        else:
            for item in self._mappingOfRemoteDispatchersAndInstances.values():
                instances = item.instances

                for instance in instances:
                    if instance.identifier == identifier:
                        ret.description = instance
                        break

        ret.dispatcher = self.getManagerOfInstance(identifier)

        return ret

    def removeDispatcher(self, dispatcher: str):
        """  Helper to remove a dispatcher.

        Args:
            dispatcher (str): The Id of the Dispatcher
        """
        if self._mappingOfRemoteDispatchersAndInstances.pop(dispatcher, False):
            self.instances.update()

    async def registerConstructor(self, identifier: str, cb):
        """ Registers a Constructor, that enables other NopeInstanceManagers to create an instance of the given type. Therefore a callback "cb" is registered with the given "typeIdentifier"

        Args:
            identifier (str): The identifier for the Constructor (Like a service)
            cb (function): The callback used, to create an instance. The Callback receives the following parameters (NopeCore, identifier:str)
        """

        if self._logger:
            self._logger.debug(
                'Adding instance generator for "' + (identifier +
                                                     '" to external Generators. Other Elements can now create instances of self type.'
                                                     ))

        async def createInstance(data):

            # Check if an instance exists or not.
            # if not => create an instance an store it.

            if data.identifier not in self._instances:
                hashableData = [data.identifier, data.params, data.type]
                try:
                    hashed = hash(hashableData)
                except BaseException as E:
                    hashed = json.dumps(hashableData)

                # It might happen, that an instance is requested multiple times.
                # therefore we have to make shure, we wont create them multiple times:
                # We will test it by using the "_internalInstances" set

                if data.identifier not in self._initializingInstance:

                    try:

                        # Mark the Instance as available.
                        self._initializingInstance[data.identifier] = hashed

                        # Create the Instance
                        _instance = await cb(self._core, data.identifier)
                        _instance.identifier = data.identifier

                        # Make shure the Data is expressed as Array.
                        if not isIterable(data.params):
                            data.params = [data.params]

                        # Initialize the instance with the parameters.
                        await _instance.init(*data.params)

                        async def disposeInstance(_data):
                            """ A Function is registered, taking care of removing
                                an instances, if it isnt needed any more.

                            Args:
                                _data (msg): The message containing the dispatcher id.
                            """

                            _data = ensureDottedAccess(_data)

                            if self._instances.get(data.identifier).usedBy:
                                try:

                                    if _data.dispatcherId in self._instances.get(
                                            data.identifier).usedBy:
                                        # Pop the dispatcher if it is present:
                                        idx = self._instances.get(
                                            data.identifier).usedBy.index(
                                            _data.dispatcherId)
                                        self._instances.get(
                                            data.identifier).usedBy.pop(idx)

                                    if len(self._instances.get(
                                            data.identifier).usedBy) == 0:
                                        # Unmark as internal instance
                                        self._internalInstances.remove(
                                            data.identifier)
                                        # Remove the Instance.
                                        await _instance.dispose()
                                        # Removes the instances
                                        self._instances.pop(data.identifier)
                                        # Remove the Function itself
                                        await self._rpcManager.unregisterService(self.getServiceName(data.identifier, 'dispose'))
                                        # Emit the instances again
                                        await self._sendAvailableInstances()

                                except ValueError:
                                    pass

                        # A Function is registered, taking care of removing
                        # an instances, if it isnt needed any more.
                        await self._rpcManager.registerService(
                            disposeInstance,
                            ensureDottedAccess({
                                'id': self.getServiceName(data.identifier, 'dispose'),
                                'schema': ensureDottedAccess({
                                    'description': f'Service, which will destructor for the instance "{data.identifier}". This function will be called internal only.',
                                    'type': 'function'})
                            })
                        )

                        # Store the Instance.
                        self._instances[data.identifier] = ensureDottedAccess({
                            'instance': _instance,
                            'usedBy': [data.dispatcherId]
                        })

                        self._internalInstances.add(data.identifier)

                        # Update the available instances:
                        await self._sendAvailableInstances()

                        # Make shure, we remove this instance.hash
                        self._initializingInstance.pop(data.identifier)

                    except BaseException as E:
                        # Make shure, we remove this instance.hash
                        self._initializingInstance.pop(data.identifier)

                        raise E

                elif self._initializingInstance.get(data.identifier) != hashed:
                    raise Exception(
                        'Providing different Parameters for the same Identifier'
                    )
                else:
                    # Check if the Instance is ready.
                    firstHint = True

                    def checker():
                        nonlocal firstHint
                        if firstHint:
                            self._logger.warn(
                                f'Parallel request for the same Instance "{data.identifier}" => Waiting until the Instance has been initialized')
                            firstHint = False
                        return data.identifier in self._instances

                    await waitFor(
                        checker,
                        ensureDottedAccess({
                            'testFirst': True,
                            'delay': 100
                        })
                    )
            else:
                # If an Element exists => Add the Element.
                self._instances.get(data.identifier).usedBy.append(
                    data.dispatcherId)

            # Define the Response.
            response = ensureDottedAccess({
                'description': self._instances.get(data.identifier).instance.toDescription(),
                'type': data.type
            })

            # Send the Response
            return response

        _cb = await self._rpcManager.registerService(
            createInstance,
            ensureDottedAccess({
                # We will add the Name to our service.
                'id': self.getServiceName(identifier, 'constructor'),
                # We dont want to have a Prefix for construcors
                'addNopeServiceIdPrefix': False,
                'schema': ensureDottedAccess({
                    'description': f'Service, which will create an construtor for the type "{identifier}".',
                    'type': 'function'
                })
            })
        )

        # Store a generator
        self._registeredConstructors[identifier] = _cb

    async def unregisterConstructor(self, identifier: str):
        """ Unregisters a present Constructor. After this, created instances are still valid, the user isnt able to create new ones.

        Args:
            identifier (str): The identifier for the Constructor (Like a service)
        """
        if identifier in self._registeredConstructors:
            if self._logger:
                self._logger.debug('Removing instance generator for "' + identifier +
                                   '" from external Generators. Other Elements cant create instances of self type anymore.')

            # We will just unregister the service from our
            # system. Therefore we just use the rpcManager

            await self._rpcManager.unregisterService(self._registeredConstructors.get(identifier))
            self._registeredConstructors.pop(identifier)

    def registerInternalWrapperGenerator(self, identifier: str, cb):
        """ Defaultly a generic wrapper will be returned, when an instance is created. you
            can specifiy specific wrapper type for different "typeIdentifier" with this method.

        Args:
            identifier (str): The identifier for the Constructor (Like a service)
            cb (function): The Callback which creates the specific wrapper.
        """
        if self._logger:
            self._logger.debug('Adding instance generator for "' + identifier +
                               '" as internal Generator. This Generator wont be used externally.')

        self._internalWrapperGenerators[identifier] = cb

    def unregisterInternalWrapperGenerator(self, identifier: str):
        """ Removes a specific generator for for a wrapper.

        Args:
            identifier (str): The identifier for the Constructor (Like a service)
        """
        if self._logger:
            self._logger.debug('Rmoving instance generator for "' + identifier +
                               '" from internal Generator. The sytem cant create elements of self type any more.')

        self._internalWrapperGenerators.pop(identifier)

    def instanceExists(self, identifier: str, externalOnly=True) -> bool:
        """  Helper, to test if an instance with the given identifier exists or not.

        Args:
            identifier (str):  identifier of the instance.
            externalOnly (bool, optional): If set to true we will only look for external instances in the external dispatchers. Defaults to True.

        Returns:
            bool: The Testresult
        """
        if identifier not in self.instances.simplified:
            return False

        if externalOnly:
            manager = self.getManagerOfInstance(identifier)
            return manager["id"] != self._id

        return True

    def getManagerOfInstance(self, identifier: str):
        """ Returns the hosting dispatcher for the given instance.

        Args:
            identifier (str): The identifier for instance (its name)

        Returns:
            INopeStatusInfo | False: The Status or false if not present.
        """
        # First we will check if the instance is available internally
        if identifier in self._internalInstances:
            return self._connectivityManager.info

        # If that isnt the case, we will check all dispatchers and search the
        # instance.
        for iter_item in self._mappingOfRemoteDispatchersAndInstances.items():
            dispatcher = iter_item[0]
            msg = iter_item[1]
            for instance in msg.instances:
                if instance.identifier == identifier:
                    return self._connectivityManager.getStatus(dispatcher)

        return None

    def getInstanceDescription(self, instanceIdentifier: str):
        """ Returns the instance Description for a specific instance. It is just a simplified wrapper
            for the "instances"-property.

        Args:
            instanceIdentifier (str): The identifier for instance (its name)

        Returns:
            INopeModuleDescription | False: The Description or False if not found.
        """
        if instanceIdentifier in self._instances:
            return self._instances.get(
                instanceIdentifier).instance.toDescription()

        for data in self._mappingOfRemoteDispatchersAndInstances.values():
            instances = data.get("instances", [])
            for instance in instances:
                if instance["identifier"] == instanceIdentifier:
                    return instance

        return False

    def constructorExists(self, typeIdentifier: str) -> bool:
        """ Helper to test if a constructor linkt to the provided "typeIdentifier" exists or not.

        Args:
            typeIdentifier (str): The identifier for the Constructor (Like a service)

        Returns:
            bool: _description_
        """
        return typeIdentifier in self.constructors.data.getContent()

    async def createInstance(self, description, options=None):
        """ Allows to create an instance. This might be the case on remote dispatchers or
            on the same element. Only a wrapper is returned, which communicates with a
            dispatcher, because we dont know where the element is provided. You can use the
            method "getDispatcherForInstance" to determine the dispatcher running the instance.

            The returned wrapper acts like a normal "internal" class.

        Args:
            description (dict-like): Description of the instance to be created
            options (dict-like, optional): Options used during creating the instance.. Defaults to None.


        Returns:
            NopeGenericModule | Registered Wrapper: An Generic Nope Module as Wrapper or a custom wrapper for the class.
        """

        # Define the Default Description
        # which will lead to an error.

        options = ensureDottedAccess(options)
        description = ensureDottedAccess(description)

        # Assign the provided Description
        _description = ensureDottedAccess({
            'dispatcherId': self._id,
            'identifier': 'error',
            'params': [],
            'type': 'unkown'
        })
        _description.update(description)
        _description.update({'dispatcherId': self._id})

        # Check if the Description is complete
        if (_description.type == 'unkown' or _description.identifier) == 'error':
            raise Exception(
                'Please Provide at least a "type" and "identifier" in the paremeters')

        # Use the varified Name (removes the invalid chars.)
        _description.identifier = varifyPath(
            _description.identifier) if self.options.forceUsingValidVarNames else _description.identifier
        if self._logger:
            self._logger.debug('Requesting an Instance of type: "' + _description.type +
                               '" with the identifier: "' + str(_description.identifier) + '"')

        try:
            _type = _description.type
            if _type not in self._internalWrapperGenerators:
                _type = '*'

            if not self.constructorExists(_description.type):
                # No default type is present for a remote
                # => assing the default type which is "*""
                raise Exception('Generator "' + _description.type +
                                '" isnt present in the network!')
            if _type in self._internalWrapperGenerators:
                if self._logger:
                    self._logger.debug('No instance with the identifiert: "' + str(_description.identifier) +
                                       '" found, but an internal generator is available. Using the internal one for creating the instance and requesting the "real" instance externally')

                # Now test if there is allready an instance with this name and type.
                # If so, we check if we have the correct type etc. Additionally we
                # try to extract its dispatcher-id and will use that as selector
                # to allow the function be called.

                _instanceDetails = self._getInstanceInfo(
                    _description.identifier)

                usedDispatcher = None

                if _instanceDetails is not None and _instanceDetails.description.type != _description.type:
                    raise Exception(
                        "There exists an Instance named: '" + str(_description.identifier) + "' but it uses a different type. Requested type: '" +
                        _description.type + "', given type: '" + str(_instanceDetails.description.type) + "'")

                elif _instanceDetails is not None:
                    usedDispatcher = _instanceDetails.dispatcher.id

                if usedDispatcher and options.assignmentValid:

                    # If we have an dispatcher, which was been used to create the instance,
                    # we have to check, the selected Dispatcher Matches our
                    # criteria.

                    if not await options.assignmentValid(_instanceDetails.description, _instanceDetails.dispatcher):
                        raise Exception('Assignment is invalid.')

                definedInstance = await self._rpcManager.performCall(
                    self.getServiceName(_description.type, 'constructor'),
                    [
                        _description
                    ],
                    options
                )

                if self._logger:
                    self._logger.debug(
                        f'Received a description for the instance "{definedInstance.description.identifier}"')

                # Create the Wrapper for our instance.
                wrapper = await self._internalWrapperGenerators.get(_type)(self._core, definedInstance.description)
                if self._logger:
                    self._logger.debug(
                        f'Created a Wrapper for the instance "{definedInstance.description.identifier}"')

                originalDispose = wrapper.dispose

                async def dispose():
                    await self.deleteInstance(wrapper.identifier)

                    await originalDispose()

                setattr(wrapper, "dispose", dispose)

                self._instances[_description.identifier] = ensureDottedAccess({
                    'instance': wrapper,
                    'usedBy': [
                        _description.dispatcherId
                    ]
                }
                )

                return wrapper

            raise Exception('No internal generator Available!')

        except Exception as e:

            if self._logger:
                self._logger.error(
                    'During creating an Instance, the following error Occurd')
                self._logger.error(formatException(e))

            raise e

    async def generateWrapper(self, description):
        # Define the Default Description
        # which will lead to an error.
        description = ensureDottedAccess(description)

        # Assign the provided Description
        _description = ensureDottedAccess({
            'dispatcherId': self._id,
            'identifier': 'error',
            'params': [],
            'type': 'unkown'
        })
        _description.update(description)
        _description.update({'dispatcherId': self._id})

        # Check if the Description is complete
        if (_description.type == 'unkown' or _description.identifier) == 'error':
            raise Exception(
                'Please Provide at least a "type" and "identifier" in the paremeters')

        # Use the varified Name (removes the invalid chars.)
        _description.identifier = varifyPath(
            _description.identifier) if self.options.forceUsingValidVarNames else _description.identifier
        if self._logger:
            self._logger.debug('Requesting an Instance of type: "' + _description.type +
                               '" with the identifier: "' + str(_description.identifier) + '"')

        try:
            _type = _description.type
            if _type not in self._internalWrapperGenerators:
                _type = '*'

            if not self.constructorExists(_description.type):
                # No default type is present for a remote
                # => assing the default type which is "*""
                raise Exception('Generator "' + _description.type +
                                '" isnt present in the network!')
            if _type in self._internalWrapperGenerators:
                if self._logger:
                    self._logger.debug('No instance with the identifiert: "' + str(_description.identifier) +
                                       '" found, but an internal generator is available. Using the internal one for creating the instance and requesting the "real" instance externally')

                # Now test if there is allready an instance with this name and type.
                # If so, we check if we have the correct type etc. Additionally we
                # try to extract its dispatcher-id and will use that as selector
                # to allow the function be called.

                _instanceDetails = self._getInstanceInfo(
                    _description.identifier)

                if _instanceDetails is not None and _instanceDetails.description.type != _description.type:
                    raise Exception(
                        "There exists an Instance named: '" + str(_description.identifier) + "' but it uses a different type. Requested type: '" +
                        _description.type + "', given type: '" + str(_instanceDetails.description.type) + "'")

                elif _instanceDetails is None:
                    raise Exception(
                        'No instance known with the idenfitier "' + str(_description.identifier) + '" !')

                definedInstance = _instanceDetails.description

                # Create the Wrapper for our instance.
                wrapper = await self._internalWrapperGenerators.get(_type)(self._core, definedInstance.description)
                if self._logger:
                    self._logger.debug(
                        f'Created a Wrapper for the instance "{definedInstance.description.identifier}"')

                originalDispose = wrapper.dispose

                async def dispose():
                    await self.deleteInstance(wrapper.indentifier)

                    await originalDispose()

                setattr(wrapper, "dispose", dispose)

                self._instances[_description.identifier] = ensureDottedAccess({
                    'instance': wrapper,
                    'usedBy': [
                        _description.dispatcherId
                    ]
                }
                )

                return wrapper

            raise Exception('No internal generator Available!')

        except Exception as e:

            if self._logger:
                self._logger.error(
                    'During creating an Instance, the following error Occurd')
                self._logger.error(formatException(e))

            raise e

    async def registerInstance(self, instance):
        """ Option, to statically register an instance, without using an specific generator etc.
            This instance is just present in the network.

        Args:
            instance (INopeInstance): The Instnce to register

        Returns:
            INopeInstance: The instance.
        """
        self._instances[instance.identifier] = ensureDottedAccess({
            'instance': instance,
            'usedBy': [],
            'manual': True
        }
        )

        self._internalInstances.add(instance.identifier)

        await self._sendAvailableInstances()

        return instance

    async def deleteInstance(self, instance, preventSendingUpdate=False) -> bool:
        """ Disposes an instance and removes it. Thereby the Instance wont be available for other
            InstanceManagers in the system.

        Args:
            instance (any): The Instance to consider
            preventSendingUpdate (bool, optional): If set to True the other systems wont be notified. This is for internal purpose only. Defaults to False.

        Returns:
            bool: The success
        """
        # Block to find the instance.
        # Based on the property (string or instance)
        # the corresponding instance object has to be select.

        _instance = None
        _identifier = None

        if isinstance(instance, str):
            _instance = self._instances.get(instance)
            _identifier = instance
        else:
            for data in self._instances.values():
                if instance == data.instance:
                    _instance = data
                    _identifier = data.instance.identifier
                    break
        try:
            params = ensureDottedAccess({
                'dispatcherId': self._id,
                'identifier': _identifier
            })

            # Call the corresponding Dispose Function for the "real" instance
            # All other elements are just accessors.
            await self._rpcManager.performCall(
                self.getServiceName(_identifier, "dispose"),
                [
                    params
                ]
            )
        except BaseException as E:
            # Only if it is an internal
            # Instance, we do not want to
            # throw that error, otherwise
            # we want that error to be
            # present.
            if _instance:
                pass
            else:
                raise E

        # if the instance has been found => delete the instance.
        if _instance:
            _instance.usedBy.pop()
            if len(_instance.usedBy) == 0:

                # Delete the Identifier
                self._instances.pop(_instance.instance.identifier)

                # Check if an update should be emitted or not.
                if not preventSendingUpdate:
                    # Update the Instances provided by this module.
                    await self._sendAvailableInstances()

            return True
        return False

    async def getInstancesOfType(self, typeToGet: str):
        """ Creates Wrappers for the Type of the given element.

        Args:
            typeToGet (str): Type of the instances to get the wrappers for.

        Returns:
            list: List containing all kown Elements of the given type.
        """

        indentifier = map(lambda item: item.identifier, filter(
            lambda item: item.type == typeToGet, self.instances.data.getContent()))

        promises = []

        for identifier in indentifier:
            promises.append(
                self.createInstance(
                    ensureDottedAccess({
                        'identifier': identifier,
                        'type': typeToGet,
                        'params': []
                    })
                )
            )

        # Wait to generate all Instances.
        if promises:
            result = await asyncio.gather(*promises)
        return result

    def reset(self):
        self._mappingOfRemoteDispatchersAndGenerators.clear()
        self._mappingOfRemoteDispatchersAndInstances.clear()
        self.constructors.update()
        self.instances.update()
        self._internalWrapperGenerators = dict()
        self._registeredConstructors = dict()

        # If Instances Exists => Delete them.
        if self._instances:

            promises = []

            # Dispose all Instances.
            for name, instance in self._instances.items():

                def onDone(p):
                    if p.exception() and self._logger:
                        self._logger.error(
                            'Failed Removing Instance "' + name + '"')
                        self._logger.error(formatException(e))

                promise: asyncio.Future = self.deleteInstance(name, True)
                promise.add_done_callback(onDone)

                if self._logger:
                    self._logger.warn('Disposing instance "' + name + '"')

                promises.append(promise)

            EXECUTOR.callParallel(asyncio.gather, *promises)

        self._instances = dict()
        self._externalInstances = dict()
        self._internalInstances = set()
        self._initializingInstance = dict()
        self._externalInstancesNames = set()
        self.internalInstances.setContent([])

        if self._communicator.connected.getContent():
            EXECUTOR.callParallel(self._sendAvailableInstances)

    async def dispose(self):
        self.reset()
        self.instances.dispose()
        self.__disposed = True

    def __del__(self):
        if not self.__disposed:
            EXECUTOR.callParallel(self.dispose)



from asyncio import Future

from ..helpers import generateId, DottedDict, Emitter, ensureDottedAccess, isAsyncFunction, getTimestamp, Promise, \
    EXECUTOR


# from ..logger.getLogger import getNopeLogger
# logger = getNopeLogger('obervable')


class NopeEventEmitter:

    def __init__(self, options=None):
        self.id = generateId()
        self.options = ensureDottedAccess({'generateTimestamp': True})
        self.options.update(ensureDottedAccess(options))
        self.setter = None
        self.getter = None
        self._subscriptions = set()
        self.disablePublishing = False
        self._emitter = Emitter()

    def emit(self, value, options=None):
        return self._emit(value, ensureDottedAccess(options))

    def _emit(self, value, options=None):

        options = ensureDottedAccess(options)

        _value = value
        if self.setter is not None:
            adapted = ensureDottedAccess(self.setter(value, options))
            if not adapted.valid:
                return False
            _value = adapted.data
        # Assign the Value (either use it directly or use the getter)
        _value = self.getter(_value) if self.getter is not None else _value
        if options.forced or (self.disablePublishing == False):
            options = self._updateSenderAndTimestamp(options)
            self._emitter.emit(data=ensureDottedAccess(
                {'value': _value, **options}))
            return self.hasSubscriptions
        return False

    def _updateSenderAndTimestamp(self, options: DottedDict):
        if not options.sender:
            options.sender = self.id
        if self.options.generateTimestamp:
            options.timestamp = options.get("timestamp", getTimestamp())
        return options

    def dispose(self):
        for _unsubscribe in self._subscriptions:
            _unsubscribe()
        self._subscriptions.clear()
        self._emitter.close()

    def subscribe(self, callback,
                  options: dict | DottedDict = DottedDict({'type': 'sync', 'mode': ['direct', 'sub', 'super']})):
        options = ensureDottedAccess(options)
        options.skipCurrent = self.options.showCurrent and not self.options.playHistory
        return self._subscribe(self._adaptCallback(callback, options))

    def _adaptCallback(self, callback, options):
        first = True

        def adaptedCallback(_data: DottedDict):
            nonlocal first

            data = _data.copy()

            if first and options.skipCurrent:
                first = False
                return

            first = False

            if data is not None:
                # Pop the value
                value = data.pop("value")
                # Now we call value, ... rest
                callback(value, data)

        return adaptedCallback

    def _subscribe(self, callback):

        return self._emitter.on(callback=callback)

    def once(self, callback, options=None):
        ret = None

        def adaptedCallback(data, rest):
            ret.unsubscribe()
            callback(data, rest)

        ret = self.subscribe(adaptedCallback, options)
        return ret

    def waitFor(self, testCallback=None, options=None):

        # Convert our Options.

        _options = ensureDottedAccess({
            'testCurrent': True,
            "timeout": 0
        })
        _options.update(ensureDottedAccess(options))

        if testCallback is None:
            # Define a Test Callback
            def test(value, *args):
                return value == True

            testCallback = test

        subscription = None
        timeout = None

        def callback(resolve, reject):
            nonlocal subscription
            nonlocal timeout

            first = True
            resolved = False

            def finish(err, sucessfull, data, timedout):
                nonlocal first
                nonlocal resolved
                nonlocal subscription
                nonlocal timeout

                if err:
                    reject(err)

                    if subscription:
                        subscription.unsubscribe()
                        subscription = None

                    if timeout and not timedout:
                        timeout.cancel()
                elif sucessfull:
                    if subscription:
                        subscription.unsubscribe()
                        subscription = None
                    if not resolved:
                        resolved = True
                        resolve(data)

                    if timeout and not timedout:
                        timeout.cancel()

            def check_data(value, rest):
                nonlocal first
                nonlocal resolved
                nonlocal subscription

                if (first and _options.testCurrent) or not first:
                    if isAsyncFunction(testCallback):
                        # Try to offload the Function
                        prom: Future = Promise.cast(testCallback(value, rest))

                        def done(p: Future):
                            nonlocal first
                            first = False

                            if p.done():
                                finish(False, True, p.result(), False)
                            if p.exception():
                                finish(p.exception(), False, False, False)

                        prom.add_done_callback(done)
                    else:
                        result = testCallback(value, rest)
                        finish(False, result, value, False)

                        first = False

            try:
                if _options.timeout > 0:
                    timeout = EXECUTOR.setTimeout(
                        finish, _options.timeout, TimeoutError("Time elapsed"), False, False, True)

                subscription = self.subscribe(check_data)
            except BaseException:
                reject(Exception("Failed to subscribe"))

        return Promise(callback)

    def waitForUpdate(self, options=DottedDict({'testCurrent': True})):
        # Define a Callback
        def callback(resolve, reject):
            def cb(value, rest):
                resolve(value)

            self.once(cb, options)

        return Promise(callback)

    @property
    def hasSubscriptions(self):
        return self._emitter.hasSubscriptions()

    @property
    def observerLength(self):
        return self._emitter.amountOfSubscriptions()



# DISCLAIMER: This lib is based on:
# https://github.com/fpom/snakes/blob/master/snakes/plugins/__init__.py

"""This package implements a plugin system. Plugins
themselves are available as modules within the package.
Examples below are based on plugin `hello` that is distributed with
NoPE to be used as an exemple of how to build a plugin. It extends
class `NopeDispatcher` adding a method `hello` that says hello displaying
the name of the net.
## Loading plugins ##
The first example shows how to load a plugin: we load
`nope.plugins.hello` and plug it into `nope.dispatcher`, which results
in a new module that is actually `nope.dispatcher` extended by
our plugin (`nope.plugins.hello`).
>>> from nope import getLayer
>>> import nope.plugins as plugins
>>> nope_extended = plugins.load('nope.plugins.hello', 'nope.dispatcher')
>>> manager = NopeRpcManager({
...     "communicator": await getLayer("event"),
...     "logger": False,
... }, lambda *args: "test", "test")
>>> manager.hello()
"Hello from the Plugin"
"""

import imp
import importlib
import inspect
import sys
from functools import wraps

from nope.logger import getNopeLogger, DEBUG

_LOGGER = getNopeLogger("plugin-system", DEBUG)

_PLUGIN_COUNTER = 0
_PLUGINS = dict()
_LOADED = list()


def _set(obj, accessor: str, value):
    """ Helper to set items.

    Args:
        obj: Object to manipulate
        accessor (str): The Attribute to change
        value (_type_): The value to use.
    """
    try:
        obj[accessor] = value
    except BaseException as E:
        try:
            setattr(obj, accessor, value)
        except BaseException as E:
            raise E


def _get(obj, accessor: str, prevent_to_load: str):
    ret = None
    additional_module = False
    try:
        ret = obj[accessor]
    except BaseException as E:
        try:
            ret = getattr(obj, accessor)
        except BaseException as E:
            raise E

    if inspect.isfunction(ret):
        # Wir haben eine funktion als attribute
        # D.h. damit wir die funktion manipulieren
        # können => wir müssen den definitions file
        # laden
        src = inspect.getsourcefile(ret)

        if src == prevent_to_load:
            # Wir laden unser addon.
            # Das wenn wir dies erneut laden ==>
            # wir können den File öfters.
            return ret, additional_module

        spec = importlib.util.spec_from_file_location(ret.__module__, src)
        ret = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(ret)

        # Wir geben jetzt noch an, dass wir
        # ein zusätzl. modul manipulieren.
        additional_module = True

    return ret, additional_module


def _rsetattr(data, path: str, value, prevent_to_load: str):
    """ Function to Set recursely a Attribute of an Object/Module

    Args:
        data (any):  The Object, where the data should be stored
        path (str): The Path of the Attribute. All are seprated by a the "."
        value (any): The Value which should be Stored in the Attribute.
    """

    obj = data

    # Wir unterteilen den Pfad in segmente
    # Über diesen wercen wir im Anschluss iterieren
    # und auf die entsprechenden items zugreifen.
    ptrs = path.split(".")
    last_path_segment = ptrs[len(ptrs) - 1]

    # Helfer, die sicherstellen wenn funktionen geladen
    # werden, dass diese Anpassungen dabei auch erfolgen.
    loaded_extra = []

    # Schleife die über die einzlenen Segmente iteriert.
    for idx, accessor in enumerate(ptrs[0:-1]):

        # Wir bekommen das submodul. Wenn der accessor
        # eine "funktion" beinhaltet -> wird dynamisch
        # das modul geladen, welches die funktion enthält.
        # Diese wird dann über "loaded_module" angezeigt
        sub, loaded_module = _get(obj, accessor, prevent_to_load)

        if loaded_module:
            # Wenn wir dyn. das modul welches die funktion
            # definiert lädt, stellen wir sicher, dass dieses
            # modul die änderung erhalten wird. später nehmen
            # wir dann die neue definition der methode und weisen
            # sie dem aktuellen modul zu.
            loaded_extra.append((obj, accessor, sub))

        # Falls wir neue elemente Anlegen => erstelle ein dotted-dict.
        # darüber können wir dann darauf zugreifen.
        if sub is None:
            # Define a dict.
            _set(obj, accessor, dict())
            # Now extract the dict.
            sub = obj[accessor]
            _LOGGER.warn(
                "Watchout. You are using the plugin-system. The code reached an untested code-sections. Please contact m.karkowski@zema.de")

        # Nun gehen wir mit dem nächsten segment des pfades
        # genau so vor.
        obj = sub

    # Wir haben unser ziel erreicht und weisen nur noch
    # das element zu.
    _set(obj, last_path_segment, value)

    # Falls wir funktionen geändert haben nehmen wir dann die neue
    # definition der methode und weisen sie den aktuellen modulen zu.
    # Sicherheitshalber machen wir dies in negativer reihenfolge.
    for o, a, v in reversed(loaded_extra):
        _set(o, a, getattr(v, a))


def _update(module, objects):
    """ Update the module based on the given adapted items. This might be

    Args:
        module: The module to update.
        objects: The changed elements.

    Raises:
        ValueError: If we cannot extend the behavior.
    """
    for obj in objects:
        if isinstance(obj, tuple):
            try:
                n, o = obj
            except BaseException:
                raise ValueError("expected (name, object) and got '%r'" % obj)
            setattr(module, n, o)
        elif inspect.isclass(obj) or inspect.isfunction(obj):
            setattr(module, obj.__name__, obj)
        elif isinstance(obj, dict):
            for k, v in obj.items():
                setattr(module, k, v)
        else:
            raise ValueError("cannot plug '%r'" % obj)


def _getOccurence(lib, lib_name: str, occurence=None, orignal_srcs=None, refs=None, func=False,
                  pre_path: str = "", max_recursion=100, current_depth=0):
    """ Goal of this function is to determine the occurence of the all items, provided in the lib.

    Args:
        lib (_type_): _description_
        ret (_type_, optional): _description_. Defaults to None.
        pre_path (str, optional): _description_. Defaults to "".

    Returns:
        _type_: _description_
    """
    if occurence is None:
        occurence = dict()
    if orignal_srcs is None:
        orignal_srcs = dict()
    if refs is None:
        refs = dict()

    if current_depth > max_recursion:
        raise Exception(
            "Max Recursion reached. Propably you import the package inside of a subpackage!")

    path_segment = pre_path + "." if pre_path else ""
    libName = lib.name if hasattr(lib, "name") else lib.__name__

    # The trick to determine if we only reexport the original stuff,
    # is to check the lib name and the path:
    if pre_path and lib_name + "." + pre_path != libName and not func:
        # We mark the item as reexport.
        refs[lib_name + "." + pre_path] = libName
    elif func:
        pass

    if not libName.startswith(lib_name):
        return occurence, orignal_srcs, refs

    l = inspect.getmembers(lib)

    for name, obj in inspect.getmembers(lib):

        if name == "getTimestamp":
            pass

        if name.startswith("_"):
            continue

        elif inspect.ismodule(obj) or inspect.isfunction(obj):

            obj_to_use = obj
            _is_func = False

            if inspect.isfunction(obj) and obj.__module__.startswith(lib_name):

                if obj.__module__ != libName:

                    # Magic:
                    # Wenn wir mit einer funktion arbeiten ->
                    # dann müssen wir das dazu gehörige module laden
                    # in dem die funktion definiert wird.

                    # Dies ist aber nur erforderlich, falls wir nicht
                    # schon das selbe modul betrachten, ansonsten werden
                    # die darin definierten Funktionen erkannt und das modul
                    # wieder geladen --> Endlos schleife. (Dies fangen wir
                    # über die If-condition ab)

                    src = inspect.getsourcefile(obj)
                    spec = importlib.util.spec_from_file_location(
                        obj.__module__, src)
                    mod = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(mod)

                    # Das modul ist nun geladen -> wir werden dieses
                    # analysieren.
                    obj_to_use = mod

                    _is_func = True

                else:
                    # Wir haben das modul schon offen -> Vergesse es.
                    obj_to_use = False

                # Wir speichern noch das vorkommen der Nutzung.
                if name not in occurence:
                    occurence[name] = set()
                occurence[name].add(pre_path)

                # Wir stellen zudem sicher, dass wir das original gespeicher
                # haben.
                orignal_srcs[name] = obj.__module__

            if not obj_to_use or inspect.isbuiltin(obj_to_use):
                # Falls es sich um ein builtin handel oder
                # wir vorher gesagt haben dass das modul
                # uninteressant ist --> Wir skippen den nächsten
                # schritt
                continue
            try:
                src = inspect.getsourcefile(obj_to_use)

                if lib_name not in src:
                    # Mit dieser Funktion fangen wir
                    # andere Libraries ab, die wir hier
                    # nicht betrachten wollen
                    continue

                # Wir haben nun eine relevante library analysiere diese
                # rekursiv.
                occurence, orignal_srcs, refs = _getOccurence(
                    obj_to_use,
                    lib_name,
                    occurence,
                    orignal_srcs,
                    refs,
                    _is_func,
                    path_segment + name,
                    max_recursion,
                    current_depth + 1
                )
            except BaseException:
                # Wenn dabei was schief läuft
                # logge den error.
                pass

        else:
            if name not in occurence:
                occurence[name] = set()
            occurence[name].add(pre_path)

            if inspect.isclass(obj) or inspect.isfunction(obj):
                orignal_srcs[name] = obj.__module__

    return occurence, orignal_srcs, refs


def _getPluginInfo(plugin):
    if isinstance(plugin, str):
        plugin = __import__(plugin, fromlist=["__name__"])

    return plugin.name, plugin.base, plugin.depends, plugin.conflicts


def install(lib, plugins, new_pkg_name: str = None):
    """ installiert ein Plugin in der Library.

    Args:
        lib (module): Die Library die betrachtet wird
        plugins (str | list<str> | func | list<func>): Die Plugins die geladen werden sollen. Siehe Beispiele.
        new_pkg_name (str, optional): Name unterdem das Update gespeicher werden soll (nur partial die Änderung). Defaults to None.

    Returns:
        (module, list(str), list(str)): Die angepasste lib (Inline edit -> wird nicht benötigt), Liste mit geupdaten Referenzen, Liste mir dem originalen Elementen.
    """
    # Damit wir nicht unseren eigene datei laden,
    # extrahieren wir diese.
    prevent_to_load = inspect.stack()[1].filename

    _LOGGER.info(f"Using plugin installed started at '{prevent_to_load}'")

    # Nun laden wir das main module was wir anpassen werden.
    if isinstance(lib, str):
        mainModule = __import__(lib, fromlist=["__name__"])
    else:
        mainModule = lib

    mainModuleName = mainModule.__name__

    # Wir ermitteln das Auftreten aller Klassen.
    occourenceDict, srcs, refs = _getOccurence(mainModule, mainModuleName)

    plugins_to_use = []
    to_be_plugged = dict()

    # Wir stellen sicher, dass wir immer eine liste von Plugins
    # betrachten.
    if not isinstance(plugins, (list, tuple)):
        plugins = [plugins]

    # In dieser schleife stellen wir sicher, dass wir immer
    # die funktion des Plugin ladens und später verwenden.
    for plugin in plugins:
        if isinstance(plugin, str):
            if plugin not in _PLUGINS:
                # Unser Plugin muss "nachgeladen werden"
                plugin = __import__(plugin, fromlist=["__name__"])

        # Basierend auf dem Typ (dynamisch / statisch)
        # speichern wir das Plugin / die Methode.
        if hasattr(plugin, "extend"):
            plugin = plugin.extend

        if not inspect.isfunction(plugin):
            raise Exception("Failed to load the Plugin")

        # Wir speicher das Plugin.
        plugins_to_use.append(plugin)

    rpr_start = "Plugins used!\n\n" + "-" * 50 + "\nPLUGIN INSTALLTION REPORT:\n" + \
                "-" * 50 + \
                f"\n\nInstalled the following plugins in '{mainModuleName}':"
    rpr_bases = "\n\nThe following source have been modified:"
    rpr_plugins = ""
    rpr_end = f"\n\nReturning modified library '{mainModuleName}'. Watchout this may change the default behavior of '{mainModuleName}'!"

    # In dieser schleife ermitteln wir nun, welche
    # Basis module für die Plugins benötigt werden.
    # Dazu unterscheiden wir in einfache Strings
    # und listen. Je nachdem laden wir die Basis module
    # und speichern diese in "to_be_plugged"
    for plugin in plugins_to_use:
        if isinstance(plugin.base, str):
            if plugin.base not in to_be_plugged:
                to_be_plugged[plugin.base] = __import__(
                    plugin.base, fromlist=["__name__"])

                rpr_plugins += f"\n\t- {plugin.name}"
                rpr_bases += f"\n\t- {plugin.base}"

            # Sicherstellen, dass wir mit einer Liste arbeiten
            plugin.base = [plugin.base]

        elif isinstance(plugin.base, (list, tuple)):
            for item in plugin.base:
                if item not in to_be_plugged:
                    to_be_plugged[item] = __import__(
                        item, fromlist=["__name__"])

                    for base in plugin.base:
                        rpr_bases += f"\n\t- {base}"
                    rpr_plugins += f"\n\t- {plugin.name}"

    changes = dict()

    # In dieser schleife nehmen wir nun die plugins und wenden diese an.
    # Durch die Anwendung dieser in einer schleife können wir verschiedene
    # Plugins kombinieren.
    for plugin in plugins_to_use:

        modules = [to_be_plugged[item] for item in plugin.base]

        mainModule, adapted_modules, changes = plugin(
            mainModule, modules, changes)

        # Jetzt spiegeln wir die Änderungen zurück in die module.
        for idx, item in enumerate(plugin.base):
            to_be_plugged[item] = adapted_modules[idx]

    updated = []
    skipped = []

    for module, changed in changes.items():

        # Wir speichern noch den Namen des originalen Element
        # Dieses darf NICHT überschrieben werden, das sonst die
        # Vererbung nicht klappt.
        destPath = module.__name__

        # Nun wissen wir was sich geändert hat und wir müssen diese
        # Änderungen im Modul implementieren:
        for name, adapted in changed.items():

            if name not in srcs:
                # Das Element ist noch nicht vorhanden (bspw. eine neue funktion)
                # daher definieren wir einen Namen zum zuweisen der Variable.
                set_path = name

                if destPath != mainModuleName:
                    set_path = destPath[len(mainModuleName) + 1:] + "." + name

                _rsetattr(mainModule, set_path, adapted, prevent_to_load)

            else:
                # We are adding an alread existing item.
                occourence = occourenceDict.get(name, set())

                for path in occourence:
                    path_to_use = mainModuleName + "." + path if path else mainModuleName
                    set_path = path + "." + name if path else name

                    # if we working with a reference like here:
                    # we skip adapting the item:
                    if path_to_use in refs:
                        skipped.append(set_path)
                        continue
                    # We have to maintain the original item.
                    elif name in srcs and path_to_use == srcs[name]:
                        skipped.append(set_path)
                        continue
                    # We already updated our destination. we dont need that.
                    # otherwise we cannot access updated plugins.
                    elif destPath == path_to_use:
                        skipped.append(set_path)
                        continue

                    # We now update the lib item.
                    _rsetattr(mainModule, set_path, adapted, prevent_to_load)
                    updated.append(set_path)

    sys.modules[mainModuleName] = mainModule

    if new_pkg_name is not None:
        to_be_plugged.__name__ = new_pkg_name
        sys.modules[new_pkg_name] = to_be_plugged
        inspect.stack()[1][0].f_globals[new_pkg_name] = to_be_plugged

    _LOGGER.warn(rpr_start + rpr_plugins + rpr_bases + rpr_end)

    return mainModule, updated, skipped


def plugin(base, name: str = None, depends=[], conflicts=[]):
    """ Helper to create a plugin. Used as decorator for the extending function.

        The base defines the bases of the module.
    Args:
        base (str, list<str>): The base that has to be adapted.
        depends (list, optional): Dependencies. Defaults to [].
        conflicts (list, optional): Conflicts. Defaults to [].
    """

    def wrapper(fun):
        @wraps(fun)
        def extend(mainModule, modules: list, changes: dict):
            try:
                loaded = set(modules.__plugins__)
            except AttributeError:
                loaded = set()
            for name in depends:
                if name not in loaded:
                    modules = install(name, modules)
                    loaded.update(modules.__plugins__)

            conf = set(conflicts) & loaded
            if len(conf) > 0:
                raise ValueError("plugin conflict (%s)" % ", ".join(conf))

            # Aufrufen des eigentlichen Plugins.
            adaptions = fun(*modules)

            if len(modules) == 1:
                adaptions = [adaptions]

            if len(modules) > 1 and len(adaptions) != len(modules):
                raise Exception(
                    "Can not assign the adaped items. Please make shure you are providing the correct length!")

            adapted_modules = list()

            for idx, changes_per_module in enumerate(adaptions):

                module = modules[idx]

                if module not in changes:
                    changes[module] = dict()

                if not isinstance(changes_per_module, (tuple, list)):
                    changes_per_module = [changes_per_module]

                for obj in changes_per_module:

                    if inspect.isclass(obj) or inspect.isfunction(obj):
                        if obj.__name__ in changes[module]:
                            raise Exception("Adapted name used twice")
                        changes[module][obj.__name__] = obj
                    elif isinstance(obj, dict):
                        for varname, v in obj.items():
                            if varname in changes[module]:
                                raise Exception("Adapted name used twice")
                            changes[module][varname] = v

                changes[module].update()

                adapted_modules.append(
                    _build(
                        fun.__module__,
                        module,
                        *changes_per_module))

            return mainModule, adapted_modules, changes

        module = sys.modules[fun.__module__]
        module.__test__ = {"extend": extend}

        extend.base = base
        extend.depends = depends
        extend.conflicts = conflicts
        extend.name = name

        # We will define a custom Pluginname if required.
        if extend.name is None:
            global _PLUGIN_COUNTER
            extend.name = f"anonymousPlugin{_PLUGIN_COUNTER}@{inspect.getsourcefile(inspect.stack()[1][0])}"
            _PLUGIN_COUNTER += 1
        else:
            _PLUGINS[extend.__name__] = extend

        _PLUGINS[extend] = extend

        return extend

    return wrapper


def _build(name, module, *objects):
    """Builds an extended module.
    The parameter `module` is exactly that taken by the function
    `extend` of a plugin. This list argument `objects` holds all the
    objects, constructed in `extend`, that are extensions of objects
    from `module`. The resulting value should be returned by `extend`.
    @param name: the name of the constructed module
    @type name: `str`
    @param module: the extended module
    @type module: `module`
    @param objects: the sub-objects
    @type objects: each is a class object
    @return: the new module
    @rtype: `module`
    """
    result = imp.new_module(name)
    result.__dict__.update(module.__dict__)
    _update(result, objects)
    result.__plugins__ = (module.__dict__.get("__plugins__",
                                              (module.__name__,))
                          + (name,))
    for obj in objects:
        if inspect.isclass(obj):
            obj.__plugins__ = result.__plugins__
    return result



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from .getPackageLoader import getPackageLoader
from .loadPackages import loadConfig, loadDesiredPackages
from .nopePackkageLoader import NopePackageLoader



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from nope.helpers import maxOfArray, minOfArray


ValidDefaultSelectors = [
    "master",
    "first",
    "dispatcher",
    "host",
    "free-ram",
    "cpu-usage",
]


def generateSelector(selector, core):
    """ A Helper Function, to generate the Basic selector Functions.

    params:
        selector (master|first|dispatcher|host|cpu-usage)

    """

    if selector == 'master':

        # Define a function, which will select the same master.
        async def masterSelector(opts):
            masterId = core.connectivityManager.master.id
            data = core.rpcManager.services.keyMappingreverse
            if opts.serviceName in data:
                arr = list(data[opts.serviceName])
                if masterId in arr:
                    return masterId
            raise Exception('No matching dispatcher present.')

        return masterSelector

    elif selector == 'first':

        async def firstFound(opts):
            data = core.rpcManager.services.keyMappingreverse
            if opts.serviceName in data:
                arr = list(data[opts.serviceName])
                if len(arr) > 0:
                    return arr[0]
            raise Exception('No matching dispatcher present.')

        return firstFound

    elif selector == 'dispatcher':

        async def ownDispatcher(opts):
            ids = core.connectivityManager.dispatchers.data.getContent()
            if core.id in ids:
                return core.id
            raise Exception('No matching dispatcher present.')

        return ownDispatcher

    elif selector == 'host':

        host = core.connectivityManager.info.host.name

        async def sameHost(opts):
            data = core.rpcManager.services.keyMappingreverse
            if opts.serviceName in data:
                items = list(data[opts.serviceName])
                hosts = list(
                    map(
                        lambda item: core.connectivityManager.dispatchers.originalData[
                            item].host.name,
                        items
                    )
                )

                if host in hosts:
                    return items[hosts.index(host)]

            raise Exception('No matching dispatcher present.')

        return sameHost

    elif selector == 'cpu-usage':

        async def cpuUsage():
            dispatchers = core.connectivityManager.dispatchers.data.getContent()
            best_option = minOfArray(dispatchers, 'host.cpu.usage')
            if best_option.index >= 0:
                return dispatchers[best_option.index]
            raise Exception('No matching dispatcher present.')

        return cpuUsage

    elif selector == 'free-ram':

        async def ramUsage():
            dispatchers = core.connectivityManager.dispatchers.data.getContent()
            best_option = maxOfArray(dispatchers, 'host.ram.free')
            if best_option.index >= 0:
                return dispatchers[best_option.index]

            raise Exception('No matching dispatcher present.')

        return ramUsage

    else:
        raise Exception('Please use a valid selector')



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de


import argparse
import asyncio
import json
import logging
import sys

from nope.loader import getPackageLoader, loadConfig, loadDesiredPackages
from nope.communication import getLayer, LAYER_DEFAULT_PARAMETERS, VALID_LAYERS
from nope.dispatcher.rpcManager.selectors import ValidDefaultSelectors
from nope.helpers import ensureDottedAccess, generateId, EXECUTOR, generateId


def getDefaultParameters():
    return {
        "file": "./config/settings.json",
        "channel": "event",
        "skipLoadingConfig": False,
        "channelParams": "not-provided",
        "log": "info",
        "dispatcherLogLevel": "info",
        "communicationLogLevel": "info",
        "delay": 2,
        "timings": {},
        "defaultSelector": "first",
        "forceUsingSelectors": False,
        "preventVarifiedNames": False,
        # "logToFile": False,
        "id": generateId(),
        "profile": False,
        "useBaseServices": True,
    }


def getArgs(add_mode=True, parser=None, default_args=None):
    """ Helper Function to extract the Arguments
    """

    if default_args is None:
        default_args = dict()

    if parser is None:
        parser = argparse.ArgumentParser(
            description='cli-tool to run the backend')
    if add_mode:
        parser.add_argument('mode', type=str, default="run",
                            help='option, used to run the code')
    parser.add_argument('--file', type=str, default="config/settings.json", dest='file',
                        help='configuration file to use. This File contains the defined packages.')
    parser.add_argument('-c', type=str, default=default_args.get("channel", "event"), dest='channel',
                        help='name of the communication layer to use. Possible values are: ' + ', '.join(
                            [key for key in LAYER_DEFAULT_PARAMETERS.keys()]))
    parser.add_argument('-p', '--channelParams', type=str, default=default_args.get("channelParams", "not-provided"), dest='channelParams',
                        help='name of the communication layer to use. Possible values are: ' + ', '.join(
                            [key for key in LAYER_DEFAULT_PARAMETERS.keys()]))
    parser.add_argument('-s', '--skip-loading-config', dest='skipLoadingConfig', action='store_true',
                        help='Skips the Configuration File.')

    parser.add_argument("--default-selector", default=default_args.get("defaultSelector", "first"), dest="defaultSelector",
                        help="The default-selector to select the service providers. Possible Values are: " +
                        ", ".join(ValidDefaultSelectors)
                        )

    parser.add_argument("--log-to-file", help="Log will be stored in a logfile.",
                        action="store_true", dest='logToFile')

    parser.add_argument('-l', '--log', type=str, default=default_args.get("log", "debug"), dest='log',
                        help='Level of the Logger. Valid values are "debug", "info"')

    parser.add_argument('--id', type=str, default=default_args.get("id", generateId(use_as_var=True, pre_string="_dispatcher")), dest='id',
                        help="Define a custom id to the Dispatcher",)

    parser.add_argument('--dispatcher-log', type=str, default=default_args.get("dispatcherLogLevel", "info"), dest='dispatcherLogLevel',
                        help='Specify the Logger Level of the Dispatcher. Defaults to "info"')

    parser.add_argument('--communication-log', type=str, default=default_args.get("communicationLogLevel", "info"), dest='communicationLogLevel',
                        help='Specify the Logger Level of the Communication. Defaults to "info"')

    parser.add_argument('--noBaseServices', default=False, dest='useBaseServices', action='store_true',
                        help='Flag to enable prevent the base Services to be loaded')

    args = parser.parse_args()

    if args.channel in [key for key in VALID_LAYERS.keys()]:
        pass
    else:
        print("invalid communication-layer has been selected")
        raise Exception("invalid communication-layer has been selected")

    if args.channelParams == "not-provided":
        args.channelParams = None

    if args.channelParams is not None:
        try:
            try:
                args.params = json.loads(args.params)
                print(args.params)
            except Exception as E:
                "".startswith
                if not (args.params.startswith(
                        "{") or args.params.startswith("[")):
                    args.params = json.loads('"' + args.params + '"')
                else:
                    print(E)
                    raise Exception("Please provide valid JSON")
        except BaseException:
            print("Please provide valid JSON")
            sys.exit()

    args.useBaseServices = not args.useBaseServices or default_args.get(
        "useBaseServices", False)
    args.skipLoadingConfig = args.skipLoadingConfig or default_args.get(
        "skipLoadingConfig", False)
    args.logToFile = args.logToFile or default_args.get("logToFile", False)

    ret = getDefaultParameters()
    ret.update(args.__dict__)
    return ensureDottedAccess(ret)


async def generateNopeBackend(_args: dict, run=False):
    """ Helper Function, which will create a Nope Instance.
    """
    args = getDefaultParameters()
    args.update(_args if isinstance(_args, dict) else _args.__dict__)

    # Test if we need to parse the Data.
    if isinstance(args, dict):
        args = ensureDottedAccess(args)

    if getattr(args, "channel", "event") in VALID_LAYERS:
        pass
    else:
        print("invalid communication-layer has been selected")
        return

    # Get a Communicator
    communicator = await getLayer(
        args["channel"],
        args.get("params", None),
        args["communicationLogLevel"]
    )

    # Get the Loader
    loader = getPackageLoader(
        communicator=communicator,
        **args
    )

    await loader.dispatcher.ready.waitFor()
    if args.delay:
        await asyncio.sleep(args.delay)

    def __run():

        try:
            if not getattr(args, "skip_loading_config", False):
                # Load all Packages
                EXECUTOR.callParallel(
                    loadDesiredPackages,
                    loader,
                    loadConfig(args.file),
                    delay=args.delay
                )

        except FileNotFoundError:
            print("Configuration-File not Found!")

        def stop():
            print("Calling Stop")
            # Now we have to run our loop,
            # until the bridge has been disposed.
            EXECUTOR.loop.run_until_complete(
                loader.dispatcher.dispose()
            )

        EXECUTOR.run(before=stop)

    if run:
        try:
            if not getattr(args, "skip_loading_config", False):
                # Load all Packages
                await loadDesiredPackages(
                    loader,
                    loadConfig(args.file).get("packages",[]),
                    delay=args.delay
                )                

        except FileNotFoundError:
            print("Configuration-File not Found!")

        connections = []

        with open(args.file, 'r') as file:
            raw_data = file.read()
            config = json.loads(raw_data)
            connections = config.get("connections",[])

        for item in connections:
            if item["name"] == "io-client" or item["name"] == "mqtt":          
                addLayer(
                    loader.dispatcher.communicator,
                    item["name"],
                    item["url"],
                    item["log"],
                    item["considerConnection"],
                    item["forwardData"]
                )

            elif "event":
                pass
            else:
                raise Exception("Using unkown Connection :(")

        try:
            await loader.dispatcher.runEndless()
        except asyncio.CancelledError:
            EXECUTOR.callParallel(loader.dispatcher.dispose)
            # await loader.dispatcher.dispose()

    return loader, loader.dispatcher, communicator, __run


async def run(add_mode, run=False):
    args = getArgs(add_mode)

    try:
        file_name = args.file

        with open(file_name, 'r') as file:
            raw_data = file.read()
            config = json.loads(raw_data)

            args = getArgs(add_mode, default_args=config["config"])
            args.file = file_name

    except BaseException as E:
        print(E)
        pass

    _, __, ___, run = await generateNopeBackend(args, run=True)


def run_cli(add_mode=True):
    args = getArgs(add_mode)

    try:
        with open(args.file, 'r') as file:
            raw_data = file.read()
            config = json.loads(raw_data)

            args = getArgs(add_mode, default_args=config["config"])

    except BaseException as E:
        print(E)
        pass

    task = asyncio.ensure_future(generateNopeBackend(args, run=True))
    try:
        EXECUTOR.start(task)
    except KeyboardInterrupt:
        print()
        # IF we detect a cancel, we dispose every thing
        task.cancel()
        EXECUTOR.stop()
        EXECUTOR.dispose()


if __name__ == "__main__":
    import sys
    import os
    import io

    SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
    sys.path.append(os.path.dirname(SCRIPT_DIR))

    run_cli(False)



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import asyncio

from nope.communication.bridge import Bridge
from nope.dispatcher.connectivityManager import NopeConnectivityManager
from nope.eventEmitter import NopeEventEmitter
from nope.helpers import generateId, ensureDottedAccess, isAsyncFunction, \
    formatException, DottedDict, SPLITCHAR, isIterable, isList, EXECUTOR
from nope.logger import defineNopeLogger
from nope.merging import DictBasedMergeData
from nope.observable import NopeObservable

_DEFAULT_RESULT = object()


class WrappedFunction:

    def __init__(self, func, _id, unregister):
        self._func = func
        self._unregister = unregister
        self.id = _id
        self.isAsync = isAsyncFunction(func)

    def __call__(self, *args, **kwarg):
        return EXECUTOR.callParallel(self._func, *args, **kwarg)


class NopeRpcManager:

    def __init__(self, options, defaultSelector, id=None,
                 connectivityManager=None):

        options = ensureDottedAccess(options)

        self.options = options
        self._defaultSelector = defaultSelector
        self._id = id
        self._communicator: Bridge = options.communicator
        self._connectivityManager: NopeConnectivityManager = connectivityManager

        self._runningInternalRequestedTasks = dict()
        self._registeredServices = dict()

        if self._id is None:
            self._id = generateId()

        if self._connectivityManager is None:
            self._connectivityManager = NopeConnectivityManager(
                options, self.id)

        self._logger = defineNopeLogger(options.logger, 'core.rpc-manager')
        self.ready = NopeObservable()
        self.ready.setContent(False)

        self.__warned = False

        self._mappingOfDispatchersAndServices = dict()
        self.services = DictBasedMergeData(
            self._mappingOfDispatchersAndServices, 'services/+', 'services/+/id')
        self.onCancelTask = NopeEventEmitter()

        if self._logger:
            self._logger.info(f'manager created id={self.id}')

        self._runningExternalRequestedTasks = dict()

        self.reset()
        EXECUTOR.callParallel(self._init)

    @property
    def id(self):
        return self._id

    def updateDispatcher(self, msg):
        self._mappingOfDispatchersAndServices[msg.dispatcher] = msg
        self.services.update()

    async def _handleExternalRequest(self, data, func: WrappedFunction | None = None):
        try:
            if not callable(func):
                if data.functionId not in self._registeredServices:
                    return
                func = self._registeredServices[data.functionId].get(
                    "func", None)

            if self._logger:
                self._logger.debug(
                    f'Dispatcher "{self.id}" received request: "{data.functionId}" -> task: "{data.taskId}"')

            if callable(func):
                # Now we check, if we have to perform test, whether
                # we are allowed to execute the task:
                if data.get("target", self.id) != self.id:
                    return

                # Define a list containing callbacks:
                cbs = []

                observer = None

                def on_cancel(reason, *args):
                    nonlocal observer
                    if reason.taskId == data.taskId:

                        for cb in cbs:
                            cb(reason)

                        observer.unsubscribe()
                        observer = None

                observer = self.onCancelTask.subscribe(on_cancel)

                # extract the arguments etc.
                # create an empty list and fill it afterwards
                args = [None] * len(data.params)
                for item in data.params:
                    args[item.idx] = item.data

                _result = _DEFAULT_RESULT

                if not func.isAsync and not self.__warned:
                    if self._logger:
                        self._logger.warn(
                            "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                        self._logger.warn(
                            "Offloading function to a separate thread.")
                    else:
                        print(
                            "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                        print("Offloading function to a separate thread.")
                    # We only want to warn the user once.
                    self.__warned = True

                resultPromise = func(*args)

                try:
                    if resultPromise is not None and getattr(
                            resultPromise, 'cancelCallback', False):
                        def _cancel_main(reason):
                            resultPromise.cancelCallback(reason)

                        cbs.append(_cancel_main)

                except Exception as error:
                    # The Cancel Function isn't available in
                    # the provided promise.
                    pass

                self._runningExternalRequestedTasks[data.taskId] = data.requestedBy

                # Wait for the Result to finish.
                _result = await resultPromise

                # Define the Result message
                result = {
                    'result': _result if _result is not _DEFAULT_RESULT else None,
                    'taskId': data.taskId,
                    'type': 'response'
                }

                if self._logger:
                    self._logger.debug(
                        'Internally executed requested Function for Task: ' + str(data['taskId']) + " - Function \"" +
                        data['functionId'] + '\". Sending result on ' + str(data['resultSink']))

                # Use the communicator to publish the result.
                await self._communicator.emit("rpcResponse", result)

        except Exception as error:

            if self._logger:
                self._logger.error(
                    f'Dispatcher "{self.id}" failed with request: "{data.taskId}"')
                self._logger.error(formatException(error))
            else:
                print(formatException(error))

            self._runningExternalRequestedTasks.pop(data.requestedBy, None)

            result = {
                'error': {
                    'error': str(error),
                    'msg': str(error)
                },
                'taskId': data.taskId,
                'type': 'response'
            }

            # Use the communicator to publish the result.
            await self._communicator.emit("rpcResponse", result)

    async def _handle_external_response(self, data):
        try:
            # Extract the Task
            task: DottedDict = self._runningInternalRequestedTasks.get(
                data.taskId, None)

            if task:

                # Based on the Result of the Remote => proceed.
                # Either throw an error or forward the result
                self._runningInternalRequestedTasks.pop(data.taskId)

                if data.error:
                    if self._logger:
                        self._logger.error(
                            "Failed with task " + str(data['taskId']))
                        self._logger.error(
                            "Reason: " + str(data['error']['msg']))
                        self._logger.exception(data['error'])

                    # Reject the Error:
                    task.future.set_exception(Exception(data['error']))

                    if "timeout" in task and task.timeout is not None:
                        # Assume, that the timeout has been
                        # defined with setTimeout
                        task.timeout.cancel()

                    return True

                else:
                    if self._logger:
                        self._logger.debug('Got sucessfull result for Task: "' + str(
                            data['taskId']) + '". Forwarding Result to Future.')

                    task.future.set_result(data.result)

                    if self._logger:
                        self._logger.debug("Forwarded result to future")

                    if "timeout" in task and task.timeout is not None:
                        task.timeout.cancel()

                    return True

        except Exception as error:
            if self._logger:
                self._logger.error(
                    "Error during handling an external response")
                self._logger.error(formatException(error))
            else:
                print(formatException(error))

        return False

    async def _sendAvailableServices(self):
        """ Function used to update the Available Services.
        """

        message = ensureDottedAccess({
            "dispatcher": self.id,
            "services": list(map(lambda item: item.options, self._registeredServices.values()))
        })

        if self._logger:
            self._logger.debug("sending available services")

        await self._communicator.emit("servicesChanged", message)

    async def _init(self):
        self.ready.setContent(False)

        await self._communicator.connected.waitFor()
        await self._connectivityManager.ready.waitFor()

        def onServicesChanged(msg):
            try:
                self.updateDispatcher(msg)
            except Exception as error:
                if self._logger:
                    self._logger.error("Failed to add the new services")
                    self._logger.error(formatException(error))
                else:
                    print(formatException(error))

        await self._communicator.on("servicesChanged", onServicesChanged)
        await self._communicator.on("rpcRequest", lambda data: EXECUTOR.callParallel(self._handleExternalRequest, data))
        await self._communicator.on("rpcResponse",
                                    lambda data: EXECUTOR.callParallel(self._handle_external_response, data))

        def on_cancelation(msg):
            if msg.dispatcher == self._id:
                self.onCancelTask.emit(msg)

        await self._communicator.on("taskCancelation", on_cancelation)

        def on_unregister(msg):
            if msg.identifier in self._registeredServices:
                self.unregisterService(msg.identifier)

        await self._communicator.on("rpcUnregister", on_unregister)

        # We are now listening on these changes!
        await self._communicator.on("bonjour", lambda *args: EXECUTOR.callParallel(self._sendAvailableServices))

        def onDispatchersChanged(changes, *args):
            if len(changes.added):
                # If there are dispatchers online,
                # We will emit our available services.
                EXECUTOR.callParallel(self._sendAvailableServices)
            if len(changes.removed):
                for rm in changes.removed:
                    self.removeDispatcher(rm)

        self._connectivityManager.dispatchers.onChange.subscribe(
            onDispatchersChanged)

        if self._logger:
            self._logger.info(f"core.rpc-manager {self._id} initialized!")

        self.ready.setContent(True)

    def removeDispatcher(self, dispatcherId: str):
        res = self._mappingOfDispatchersAndServices.pop(dispatcherId, False)
        if (res):
            self.services.update()

            # Now we need to cancel every Task of the dispatcher,
            # which isnt present any more.
            EXECUTOR.callParallel(self.cancelRunningTasksOfDispatcher, dispatcherId, Exception(
                "Dispatcher has been removed! Tasks cannot be executed any more."))
            # Stop executing the requested Tasks.
            EXECUTOR.callParallel(self.cancelRequestedTasksOfDispatcher, dispatcherId, Exception(
                "Dispatcher has been removed! Tasks are not required any more."))

    async def cancelTask(self, taskId: str, reason, quiet=False):
        if taskId in self._runningInternalRequestedTasks:
            task = self._runningInternalRequestedTasks[taskId]

            # Delete the task
            self._runningInternalRequestedTasks.pop(taskId)

            # Propagate the Cancellation (internally):
            task.future.set_exception(reason)

            # Propagate the Cancellation externally.
            # Therefore use the desired Mode.
            await self._communicator.emit("taskCancelation", ensureDottedAccess({
                "dispatcher": self._id,
                "reason": str(reason),
                "taskId": taskId,
                "quiet": quiet
            }))

            return True
        # Task hasnt been found => Cancel the Task.
        return False

    async def _cancelHelper(self, tasksToCancel: set, reason):
        """ Helper Function, used to close all tasks with a specific service.
        """
        if len(tasksToCancel) > 0:
            for taskId in tasksToCancel:
                await self.cancelTask(taskId, reason)

    async def cancelRunningTasksOfService(self, serviceName: str, reason):
        """ Helper Function, used to close all tasks with a specific service.
        """
        # Set containing all Tasks, that has to be canceled
        toCancel = set()

        # Filter all Tasks that should be canceled.
        for taskId, task in self._runningInternalRequestedTasks.items():
            if task.serviceName == serviceName:
                toCancel.add(taskId)

        return await self._cancelHelper(toCancel, reason)

    async def cancelRequestedTasksOfDispatcher(self, dispatcher: str, reason):
        """ Helper to cancel all Tasks which have been requested by a Dispatcher.
        """
        # Set containing all Tasks, that has to be canceled
        toCancel = set()

        for taskId, requestedBy in self._runningExternalRequestedTasks.items():
            if requestedBy == dispatcher:
                toCancel.add(taskId)

        return await self._cancelHelper(toCancel, reason)

    async def cancelRunningTasksOfDispatcher(self, dispatcher: str, reason):
        """ Cancels all Tasks of the given Dispatcher
        """
        toCancel = set()
        for taskId, target in self._runningExternalRequestedTasks.items():
            if target == dispatcher:
                toCancel.add(taskId)

        return await self._cancelHelper(toCancel, reason)

    def serviceExists(self, serviceName: str):
        """ Function to test if a specific Service exists.
        """
        return serviceName in self.services.amountOf and self.services.amountOf[
            serviceName] > 0

    def _getServiceName(self, _id: str, _type: str):
        if _type in ("request", "response"):
            return _id if _id.startswith(f"{_type}/") else f"{_type}/{_id}"
        raise Exception(
            "For the parameter 'type' the values: 'request' and 'response' are allowed!")

    async def unregisterService(self, func):
        idOfFunc: str = ""
        if isinstance(func, str):
            if self.options.forceUsingValidVarNames:
                idOfFunc = varifyPath(func)
            else:
                idOfFunc = func
        else:
            idOfFunc = func.id

        promise = self._sendAvailableServices()

        if self._logger:
            self._logger.debug(
                f'Dispatcher "{self._id}" unregistered: "{idOfFunc}"')

        ret = self._registeredServices.pop(idOfFunc, False)

        await promise

        return ret != False

    def _adaptServiceId(self, serviceName: str):
        if serviceName.startswith(f'nope{SPLITCHAR}service{SPLITCHAR}'):
            return serviceName
        return f'nope{SPLITCHAR}service{SPLITCHAR}{serviceName}'

    async def registerService(self, func, options):
        options = ensureDottedAccess(options)

        # Define / Use the ID of the Function.
        idOfFunc: str = options.id

        if idOfFunc is None:
            idOfFunc = generateId()

        if self.options.addNopeServiceIdPrefix:
            idOfFunc = self._adaptServiceId(idOfFunc)
        if self.options.forceUsingValidVarNames:
            idOfFunc = varifyPath(idOfFunc)

        options.id = idOfFunc

        if not self.__warned and not isAsyncFunction(func):
            if self._logger:
                self._logger.warn(
                    "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                self._logger.warn(
                    f'The service "{idOfFunc}" is synchronous!')
            else:
                print(
                    "!!! You have provided synchronous functions. They may break NoPE. Use them with care !!!")
                print(f'The service "{idOfFunc}" is synchronous!')
            # We only want to warn the user once.
            self.__warned = True

        # Unregister the Function:
        def unregister():
            self.unregisterService(idOfFunc)

        # Create a Wrapper
        wrapped = WrappedFunction(func, idOfFunc, unregister)

        self._registeredServices[idOfFunc] = ensureDottedAccess({
            "options": options,
            "func": wrapped
        })

        await self._sendAvailableServices()

        if self._logger:
            self._logger.debug(
                f'Dispatcher "{self._id}" registered: "{idOfFunc}"')

        return wrapped

    async def _performCall(self, serviceName, params, options=None):
        optionsToUse = ensureDottedAccess({
            "resultSink": self._getServiceName(serviceName, "response"),
            "waitForResult": True
        })
        optionsToUse.update(ensureDottedAccess(options))

        taskId = generateId()

        # Create a Future of the Loop.
        future = EXECUTOR.generatePromise(taskId=taskId)

        def clear():
            if taskId in self._runningInternalRequestedTasks:
                task = self._runningInternalRequestedTasks[taskId]

                if task.timeout is not None:
                    task.timeout.cancel()

                self._runningInternalRequestedTasks.pop(taskId)

            if self._logger:
                self._logger.debug('Clearing Callbacks from ' + str(taskId))

        try:
            tastRequest = ensureDottedAccess({
                "future": future,
                "clear": clear,
                'serviceName': serviceName,
                'timeout': None,
            })

            # Store the Future as Task.
            self._runningInternalRequestedTasks[taskId] = tastRequest

            # Define the packet to send:
            packet = {
                'functionId': serviceName,
                'params': [],
                'taskId': taskId,
                'resultSink': optionsToUse['resultSink'],
                'requestedBy': self._id,
                'target': None
            }

            # Iterate over all Parameters and
            # Determin Callbacks. Based on the Parameter-
            # Type assign it either to packet.params (
            # for parsable Parameters) and packet.callbacks
            # (for callback Parameters)
            for idx, param in enumerate(params):
                packet['params'].append({
                    'idx': idx,
                    'data': param
                })

            if not self.serviceExists(serviceName):
                error = Exception(
                    f'No Service Provider known for "{serviceName}"')
                if self._logger:
                    self._logger.error(
                        f'No Service Provider known for "{serviceName}"')
                    self._logger.error(formatException(error))

                raise error

            if self.options.forceUsingSelectors or self.services.amountOf.get(
                    serviceName, 0) > 1:

                optionsForSelector = ensureDottedAccess({
                    "rpcManager": self,
                    "serviceName": serviceName
                })

                if isinstance(optionsToUse.target, str):
                    tastRequest.target = optionsToUse.target
                elif callable(optionsToUse.selector):
                    tastRequest.target = await options.selector(optionsForSelector)
                elif isinstance(optionsToUse.selector, str):
                    tastRequest.target = await self._defaultSelector(optionsForSelector)

            else:
                tastRequest.target = list(
                    self.services.keyMappingreverse[serviceName])[0]

            packet["target"] = tastRequest.target

            await self._communicator.emit("rpcRequest", packet)

            if self._logger:
                self._logger.debug(
                    f'Dispatcher "{self._id}" putting task "{taskId}" on: "{self._getServiceName(tastRequest.serviceName, "request")}"')

            if optionsToUse.get("timeout", 0) > 0:
                async def onTimeout():
                    await self.cancelTask(
                        taskId,
                        TimeoutError(
                            f"TIMEOUT. The Service allowed execution time of {optionsToUse.timeout} [ms] has been excided")
                    )

                # Create our timeout and store it.
                tastRequest.timeout = EXECUTOR.setTimeout(
                    onTimeout, optionsToUse.timeout)

        except Exception as err:
            if self._logger:
                self._logger.error('Something went wrong on calling')
                self._logger.exception(formatException(err))
            else:
                print(formatException(err))

            # Call the Clear Function
            clear()

            # Throw an error.
            future.set_exception(err)

        # Define a Function, which could be used for cancelation.
        def _cancelTask(reason):
            EXECUTOR.callParallel(self.cancelTask, taskId, reason)

        future.cancelCallback = _cancelTask

        if not optionsToUse.waitForResult:
            EXECUTOR.loop.create_task(future)
            return future

        return await future

    async def performCall(self, serviceName, params, options=None):
        if isList(serviceName):
            if isList(options) and len(serviceName) != len(options):
                raise Exception(
                    "The length of the provided services and options does not match")

            optionsToUse = [options] * \
                len(serviceName) if not isIterable(options) else options

            futures = []

            for idx, srv in enumerate(serviceName):
                futures.append(
                    self._performCall(
                        srv,
                        params,
                        optionsToUse[idx]
                    )
                )

            return await asyncio.gather(*futures)

        else:
            if isList(options):
                raise Exception(
                    "The length of the provided services and options does not match")

            return await self._performCall(serviceName, params, options)

    def clearTasks(self):
        self._runningInternalRequestedTasks.clear()

    def unregisterAll(self, log=False):
        toUnregister = list(self._registeredServices.keys())
        for srv in toUnregister:
            EXECUTOR.callParallel(self.unregisterService, srv)
            # Log the removing of the service.
            if log and self._logger:
                self._logger.warn('removing service "' + srv + '"')
        self._registeredServices.clear()

    def reset(self):
        self.clearTasks()
        self.unregisterAll()
        EXECUTOR.callParallel(self._sendAvailableServices)

    async def dispose(self):
        self.clearTasks()
        self.ready.dispose()
        self.unregisterAll(log=True)



from asyncio import sleep

import nope
from nope import getLayer, EXECUTOR, getDispatcher
from nope.plugins import install, plugin

"""
This is an Example how to extend the behavior of any class inside of Nope.

Therefore we define a function (extend) which receives one Parameter, the
module which is located under the given path. This module must contain
the loaded class or function. We have to decorate the function with the
’plugin' decorator (which receives the path as parameter).

Inside of the function we are able to access the original class using the
`module.*ACCESSOR*`. In the example below we create an extra function
"hello_dynamic" which will just give us a simple hello world response.

Additional you are able to "store" functions into the module (see our
extend function).

If you want to manipulate specific vars --> please use a dict. for the
key use the name of the variable you want to edit.
"""


@plugin("nope.dispatcher.rpcManager")
def extend_1(module):
    "Extends `module` - rpcManager"

    class NopeRpcManager(module.NopeRpcManager):
        async def performCall(self, *args, **kwargs):
            """ Extend the original behavior """
            print("calling perform call - extend 1")
            return await module.NopeRpcManager.performCall(self, *args, **kwargs)

        def hello_dynamic(self):
            "Ask the plugin to say hello"
            print("Hello from the Dynamic Plugin")

    def hello_dynamic():
        "Ask the plugin to say hello"
        print("Hello from the Dynamic Plugin")

    hello_dynamic_vars = {"hello": "Hello from the Dynamic Plugin"}
    return NopeRpcManager, hello_dynamic, hello_dynamic_vars


# Now install our plugin
install(nope, extend_1)


# The following main is just for clearification


async def main():
    # Create our dispatcher
    dispatcher = getDispatcher({
        "communicator": await getLayer("event"),
        "logger": False,
    })

    manager = dispatcher.rpcManager

    async def hello_srv(greetings):
        print("Hello", greetings, "!")

    await dispatcher.ready.waitFor()
    await dispatcher.rpcManager.registerService(hello_srv, {"id": "hello_srv"})

    await sleep(0.1)

    manager.hello_dynamic()
    await manager.performCall("hello_srv", ["reader"])

    # Now we are able to access our variable
    print(nope.dispatcher.rpcManager.hello)

    # and our new function, we injected into the module.
    nope.dispatcher.rpcManager.hello_dynamic()


EXECUTOR.loop.run_until_complete(main())



from . import (communication, decorators, dispatcher, eventEmitter, helpers, demo, loader,

               logger, merging, observable, pubSub, plugins, modules, types)
from .cli import *
from .communication import *
from .decorators import *
from .dispatcher import *
from .eventEmitter import *
from .helpers import *
from .loader import *
from .logger import *
from .merging import *
from .modules import *
from .observable import *
from .plugins import install
from .pubSub import *
from .types import *



# This is a reference Implementation for a Plugin:
class AbstractBridgePlugin:

    def __init__(self):
        self.options = ensureDottedAccess(dict(), False)
        self._id = generateId()

        # Must contain an obervalbe with the flag ready.
        self.ready = NopeObservable()
        self.ready.setContent(False)

    @property
    def id(self) -> str:
        return self._id

    async def transformData(self, eventName: str, data):
        # Must return the following data:
        # 1. eventName: str
        # 2. data: any
        # 3.

        # Example 1:
        #   return eventName, data, None

        # Example 2:
        #   return eventName, data, None
        raise Exception("Not Implemented")

    async def init(self) -> None:
        # Expected after calling init, the ready flag will be
        # called.

        # Example:
        #   self.ready.setContent(True)

        raise Exception("Not Implemented")

    def reset(self) -> None:
        # No return type is being used.
        raise Exception("Not Implemented")

    async def dispose(self, quiet=False):
        pass



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from nope.dispatcher.baseServices import addAllBaseServices
from nope.dispatcher.nopeDispatcher import NopeDispatcher
from nope.helpers import ensureDottedAccess, EXECUTOR
from nope.modules import NopeGenericModule


def _performInit(dispatcher: NopeDispatcher, options):
    """ Helper Function to initiaize the Dispatcher.

    Args:
        dispatcher (NopeDispatcher): The Dispatcher to initalize
        dispatcherOptions (dict-like): The Options of the Dispatcher.
    """

    async def creator(core, description):
        mod = NopeGenericModule(core)
        await mod.fromDescription(description, "overwrite")
        return mod

    # Register a default instance generator:
    # Defaultly generate a NopeGenericModule
    dispatcher.instanceManager.registerInternalWrapperGenerator(
        "*",
        creator
    )

    if (options.useBaseServices):
        async def addServices():
            services = await addAllBaseServices(dispatcher)
            setattr(dispatcher, "baseServices", services)

        EXECUTOR.callParallel(addServices)

    return dispatcher


class NopeDispatcherSingleton(object):
    _instance = None

    def __new__(cls, options, id=None):
        if cls._instance is None:
            # Assign a Class
            cls._instance = _performInit(
                NopeDispatcher(options, id=id), options)

            # Put any initialization here.
        return cls._instance


def getDispatcher(
        dispatcherOptions,
        options=None):
    """Function, which woll return a nope-dispatcher
    """

    options = ensureDottedAccess(options, useNoneAsDefaultValue=True)
    dispatcherOptions = ensureDottedAccess(dispatcherOptions)

    # id = dispatcherOptions.communicator.id
    id = None

    dispatcherOptions["logger"] = dispatcherOptions["dispatcherLogLevel"]

    if options.singleton:
        return NopeDispatcherSingleton(dispatcherOptions, id)

    # Create the Dispatcher Instance.
    return _performInit(NopeDispatcher(dispatcherOptions, id=id), options)



import os
import pathlib


def createFile(path: str, with_file=True):
    """ Helper to create a path to a file.

    Args:
        path (str): The Path with or without a file.
        with_file (bool, optional): If the file is contained the Flag must be set to true. Defaults to True.
    """

    sep = os.sep
    abs_path = pathlib.Path(path).absolute()
    if not abs_path.exists():

        if with_file:
            abs_path_as_str = str(abs_path)
            path_segments = abs_path_as_str.split(sep)

            if len(path_segments) > 0:
                pathlib.Path(
                    sep.join(path_segments[:-1])).mkdir(parents=True, exist_ok=True)

            abs_path.touch(exist_ok=True)

        else:

            abs_path.mkdir(parents=True, exist_ok=True)

    return str(abs_path)



from asyncio import sleep

import nope
from nope.plugins import install, plugin

"""
You can although combine different plugins for 1 element.

Therefore the install method will receive instead of a plugin
a list of plugins. The tool will apply the installation using
the provided order.

"""


@plugin("nope.dispatcher.rpcManager")
def extend_1(module):
    "Extends `module` - rpcManager"

    class NopeRpcManager(module.NopeRpcManager):
        async def performCall(self, *args, **kwargs):
            """ Extend the original behavior """
            print("calling perform call - extend 1")
            return await module.NopeRpcManager.performCall(self, *args, **kwargs)

        def hello_dynamic(self):
            "Ask the plugin to say hello"
            print("Hello from the Dynamic Plugin")

    def hello_dynamic():
        "Ask the plugin to say hello"
        print("Hello from the Dynamic Plugin")

    hello_dynamic_vars = {"hello": "Hello from the Dynamic Plugin"}
    return NopeRpcManager, hello_dynamic, hello_dynamic_vars


@plugin("nope.dispatcher.nopeDispatcher")
def extend_2(module):
    "Extends `module` - NopeDispatcher"

    class NopeDispatcher(module.NopeDispatcher):
        async def performCall(self, *args, **kwargs):
            """ Extend the original behavior """
            print("calling perform call - from ")
            try:
                return await self.rpcManager.performCall(self, *args, **kwargs)
            except BaseException:
                print("Failed but we got the error")

    return NopeDispatcher


"""
In our case the plugin 1 is loaded first and then the second (extend_2)
In the method 'performCall' of the 'NopeRpcManager' this results in loading
the latest plugin first (extend_2) accessing the 'module.NopeRpcManager' inside
of 'extend_2' returns you the already modified 'NopeRpcManager' with plugin 1
(extend 1)

If we call the method
>>> await manager.performCall("hello_srv", ["reader"])
calling perform call - extend 1
Failed but we got the error
"""

nope, updated, skipped = install(nope, extend_1)
nope, updated, skipped = install(nope, extend_2)

print(updated, skipped)


# The following main is just for clearification


async def main():
    # Create our dispatcher
    dispatcher = nope.getDispatcher({
        "communicator": await nope.getLayer("event"),
        "logger": False,
    })

    manager = dispatcher.rpcManager

    async def hello_srv(greetings):
        print("Hello", greetings, "!")

    await dispatcher.ready.waitFor()
    await dispatcher.rpcManager.registerService(hello_srv, {"id": "hello_srv"})

    await sleep(0.1)

    # manager.hello_dynamic()
    await manager.performCall("hello_srv", ["reader"])
    await dispatcher.performCall("hello_srv", ["reader"])
    # If we call the method
    # >>> await dispatcher.performCall("hello_srv", ["reader"])
    # calling perform call - extend 2
    # calling perform call - extend 1
    # Failed but we got the error

    # Now we are able to access our variable

    # and our new function, we injected into the module.
    nope.dispatcher.rpcManager.hello_dynamic()


nope.EXECUTOR.loop.run_until_complete(main())



from nope.communication.bridge import Bridge
from nope.communication.layers import MQTTLayer, IoSocketClientLayer
from nope.helpers import generateId, DottedDict

VALID_LAYERS = DottedDict({
    'event': Bridge,
    'io-client': IoSocketClientLayer,
    'mqtt': MQTTLayer
})

LAYER_DEFAULT_PARAMETERS = DottedDict({
    'io-client': 'http://127.0.0.1:7000',
    'mqtt': 'mqtt://localhost:1883'
})


async def addLayer(bridge: Bridge, layer: str, parameter=None, logger=False, considerConnection = False, forwardData = False):
    """ Adds a Layer to the Bridge.

    Args:
        bridge (Bridge): The Bridge used to add the layer to
        layer (str): Name of the Layer. Must match the VALID_LAYERS elements
        parameter (Any, optional): The Parameter used for the Layer. Defaults to None.
        logger (bool, optional): A Definition for a Logger for the Layer. Defaults to False.
        considerConnection (bool, optional): Flag to consinder that connection for the connected flag. Defaults to False.
        forwardData (bool, optional): Enables or Disables the forwarding of the data. Defaults to False.
    """
    params = parameter if parameter is not None else LAYER_DEFAULT_PARAMETERS.get(layer, False)
    
    if layer == 'event':
        pass
    elif layer == 'io-client':
        await bridge.addCommunicationLayer(IoSocketClientLayer(params, logger), forwardData, considerConnection)
    elif layer == 'mqtt':
        await bridge.addCommunicationLayer(MQTTLayer(params, logger), forwardData, considerConnection)
    else:
        raise Exception("Wrong Layer provided!")

    bridge.connected.forcePublish()